"""Observation Class

Hosts the reflectance sampled by a given instrument of a given material
collection.

Provides methods for computing RMSE of observation compared to original
spectral library data.

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 29-04-2021
"""
import os
from pathlib import Path
from shutil import rmtree
import time
from typing import List, Union
from matplotlib import pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from scipy import interpolate
from sklearn import decomposition
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sptk.material_collection import MaterialCollection
from sptk.instrument import Instrument
from sptk.spectral_library_analyser import SpectralLibraryAnalyser
import sptk.config as cfg
from sptk.config import build_project_directory as build_pd

plt.rcParams['figure.figsize'] = cfg.FIG_SIZE

class Observation():
    """Observation Class
    Hosts the reflectance sampled by a given instrument of a given material
    collection.

    Provides methods for computing RMSE of observation compared to original
    spectral library data.
    """
    def __init__(self,
            material_collection: MaterialCollection,
            instrument: Instrument,
            load_existing: bool=cfg.LOAD_EXISTING,
            plot_profiles: bool=cfg.PLOT_PROFILES,
            export_df: bool=cfg.EXPORT_DF):
        """Constructor for Observation class

        :param material_collection: original material collection
        :type material_collection: MaterialCollection
        :param instrument: instrument that performed the sampling
        :type instrument: Instrument
        :param load_existing: instruct to use or overwrite existing directories
            and files of the same project_name, defaults to config.py setting.
        :type load_existing: bool, optional
        :param plot_profiles: plot profiles, defaults to True
        :type plot_profiles: bool, optional
        :param export_df: export DataFrames, defaults to False
        :type export_df: bool, optional
        """
        if cfg.TIME_IT:
            tic = time.perf_counter()

        p_dir, p_name = build_pd(material_collection.project_name,'observation')
        self.project_dir = p_dir
        self.project_name = p_name
        self.object_dir = Path(self.project_dir / 'observation')

        self.material_collection = material_collection
        self.categories = self.material_collection.categories

        self.instrument = instrument
        self.wvls = self.instrument.cwls().to_numpy() # set to channel cwls
        self.chnl_lbls = ['R' + str(s) for s in self.wvls] #label as R[cwl]

        if load_existing:
            existing_pkl_path = Path(self.object_dir, 'observation.pkl')
            file_exists = os.path.isfile(existing_pkl_path)
            if file_exists:
                print("Loading existing Observation DataFrame...")
                self.main_df = pd.read_pickle(existing_pkl_path)
            else:
                print("No existing DataFrame, building new Observation \
                                                for % r..." % self.project_name)
                self.main_df = self.build_new_observation(
                            material_collection,
                            instrument,
                            self.wvls)
        else:
            print("Building new Observation DataFrame  \
                                                for % r..." % self.project_name)
            self.main_df = self.build_new_observation(
                        material_collection,
                        instrument,
                        self.wvls)

        if export_df:
            self.export_main_df()
        if plot_profiles:
            self.plot_profiles()
            self.statistics()

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Observation built in {toc - tic:0.4f} seconds.")

    def __del__(self, rmdir: bool = False, rmproj: bool = False) -> None:
        """Observation Destructor object - optionally deletes
        observation directory and/or entire project directory

        :param rmdir: instruct removal of observation directory,
            defaults to False
        :type rmdir: bool, optional
        :param rmproj: instruct removal of entire project directory,
            defaults to False
        :type rmproj: bool, optional
        """
        name = self.project_name
        if rmdir:
            print("Deleting % r Observation directory..." % name)
            try:
                rmtree(Path(self.project_dir, 'observation'))
                print("% r Observation directory deleted." % name)
            except FileNotFoundError:
                print("No % r Observation directory to delete." % name)
        if rmproj:
            print("Deleting % r directory..." % name)
            try:
                rmtree(Path(self.project_dir))
                print("% r directory deleted." % name)
            except FileNotFoundError:
                print("No % r directory to delete." % name)

    @staticmethod
    def build_new_observation(
            material_collection: MaterialCollection,
            instrument: Instrument,
            cwls: List) -> pd.DataFrame:
        """Build a new Observation object by sampling the given material
        collection with the given instrument.
        """
        sample_matrix = material_collection.sample(instrument)
        # put matrix into dataframe compatible with the material collection
        main_df = pd.DataFrame(
            data=sample_matrix,
            index=material_collection.main_df.index,
            columns=cwls)
        # include the material collection header and category data
        cat_df = material_collection.get_cat_df()
        hdr_df = material_collection.get_hdr_df()
        hdr_df['Reflectance'] = '---'
        main_df = pd.concat([cat_df, hdr_df, main_df], axis=1)
        return main_df

    def add_noise(self,
            noise: float,
            n_duplicates: int,
            noise_type: str = 'shot',
            apply: bool = True) -> pd.DataFrame:
        """Add n_duplicates of noisey entries to the sampled data,
        under assumption of Gaussian distribution of noise, given by 1-sigma
        argument.

        Note: key assumption is that each filter channel was captured optimally,
            i.e. with the same count level, and thus the same shot noise, and
            that the additional dark shot noise as a function of exposure time
            has a negilible difference between filter channels. Also assumes
            that the electron count of the exposed image is >>20, which is
            valid for typical natural illumination conditions, and thus the
            Poissonian shot noise can be approximated with a Gaussian.

        :param noise: instrument noise (in reflectance units)
        :type noise: float
        :param n_duplicates: number of noisy entries to add to the data
        :type n_duplicates: int
        :param noise_type: determine if dominant noise is shot or thermal,
            defaults to shot
        :type noise_type: str
        :param apply: apply the noise to the object main_df, defaults to True
        :type apply: bool
        :return: the main dataframe
        :type return: pd.DataFrame
        """
        # access the observation dataframe and make duplicates of each entry
        obs_df = pd.concat([self.main_df]*n_duplicates).sort_index()
        # apply noise to the duplicate entries
        if noise_type == 'shot':
            sig = obs_df[self.wvls].to_numpy()*noise
            noise_array = np.random.normal(0.0, sig, obs_df[self.wvls].shape)
        elif noise_type == 'thermal':
            noise_array = np.random.normal(0.0, noise, obs_df[self.wvls].shape)
        else:
            raise ValueError(f'{noise_type} noise_type not recognised.')

        obs_df[self.wvls] = obs_df[self.wvls] + noise_array # update dataframe
        obs_df[self.wvls].clip(lower = 0.0, inplace=True) # clip to range
        # update index for unique ids
        suffix = obs_df.groupby(level=0).cumcount().astype(str).replace('0','')
        obs_df.index = obs_df.index +'.'+ suffix
        obs_df.index.name = 'Data ID'

        if apply:
            self.main_df = obs_df # update the main_df

        return obs_df

    # """
    # RMSE Computation between Observation and Material Collection
    # """

    def compute_rmse(self) -> pd.Series:
        """Compute the RMSE between the Observation and the Material Collection
        by interpolating the Observation back to the Material Collection range

        :return: RMSE of each entry
        :rtype: pd.Series
        """
        obs_refl = self.resample_wavelengths()

        mat_refl = self.material_collection.get_refl_df()
        indices_match = (mat_refl.index).equals(self.main_df.index)
        if not indices_match:
            raise IndexError("MaterialCollection and Observation Indices don't \
                                                                        match")

        mat_refl = mat_refl.to_numpy()
        rmse_data = np.sqrt(np.nanmean((mat_refl - obs_refl)**2,axis=1))

        rmse = pd.Series(data=rmse_data, index=self.main_df.index, name='RMSE')

        # append rmse to the main df in the column to the left of 'Reflectance'
        rmse_loc = self.main_df.columns.get_loc('Reflectance')
        self.main_df.insert(rmse_loc, 'RMSE', rmse)

        return rmse

    def resample_wavelengths(self) -> np.array:
        """Resample the observation data to wavelength resolution of the
        material collection.

        Note that interpolation will fill out of bound values with NaN

        :return: reflectance data interpolated to high-resolution domain.
        :rtype: np.array
        """
        obs_refl = self.get_refl_df()
        mat_col_wvls = cfg.WVLS
        finterp = interpolate.interp1d(self.wvls, obs_refl, bounds_error=False)
        obs_refl_interp = finterp(mat_col_wvls)

        return obs_refl_interp

    def get_rmse_df(self, rmse: pd.Series) -> pd.DataFrame:
        """Returns the RMSE information in a dataframe
        gives error if RMSE has not been computed yet - suggests computing the
        RMSE also gives header information

        :param rmse: RMSE of each entry
        :type rmse: pd.Series
        :return: rmse information in dataframe format
        :rtype: pd.DataFrame
        """
        hdr_df = self.material_collection.get_hdr_df()
        cat_df = self.material_collection.get_cat_df()
        labelled_rmse_df = pd.concat([cat_df, hdr_df, rmse], axis=1)
        # add category

        return labelled_rmse_df

    def export_rmse_df(self, rmse: pd.Series) -> None:
        """Export the RMSE DataFrame

        :param rmse: RMSE of each entry
        :type rmse: pd.Series
        """
        # export the RMSE dataframe
        table_path = Path(self.object_dir / 'tables')
        table_path.mkdir(parents=True, exist_ok=True)
        csv_out_file = Path(table_path, 'rmse').with_suffix('.csv')
        rmse_df = self.get_rmse_df(rmse)
        rmse_df.transpose().to_csv(csv_out_file)

    def plot_rmse(self, rmse_df: pd.DataFrame) -> None:
        """Plot the RMSE of each material entry, grouped by category and
        material type.

        :param rmse_df: RMSE results DataFrame
        :type rmse_df: pd.DataFrame
        """
        rmse_df.sort_values(by=['Category', 'Mineral Name', 'RMSE'],
                            inplace=True, ignore_index=True)
        # put index as column number
        rmse_df.reset_index(inplace=True)

        # set NaN grain sizes to 'Unspecified'
        rmse_df['Grain Size'].fillna('Unspecified', inplace=True)

        # TODO - filtering of grain size information to give numerical values.
        # Builds on new use of Grain Size in the VISOR database

        # get locations of each category
        cat_ticks = []
        cat_label_y = {}
        cat_bounds = {}
        for cat in self.categories:
            cat_lo = min(rmse_df[rmse_df.Category == cat].index)
            cat_hi = max(rmse_df[rmse_df.Category == cat].index)
            cat_bounds[cat] = [cat_lo, cat_hi]
            cat_label_y[cat] = np.mean([cat_lo, cat_hi]) + 1
            cat_ticks.append(cat_lo)

        # get locations of each Mineral Name
        min_ticks = []
        mineral_names = rmse_df['Mineral Name'].unique()
        mineral_label_y = {}
        mineral_bounds = {}
        for mineral_name in mineral_names:
            min_lo = min(rmse_df[rmse_df['Mineral Name'] == mineral_name].index)
            min_hi = max(rmse_df[rmse_df['Mineral Name'] == mineral_name].index)
            mineral_bounds[mineral_name] = [min_lo, min_hi]
            mineral_label_y[mineral_name] = np.mean([min_lo, min_hi])
            min_ticks.append(min_lo)

        height = 0.2*len(rmse_df)
        if height < 15:
            height = 15 # limit the minimum height to 15 cm
        _, ax = plt.subplots(figsize=[15*cfg.CM, height*cfg.CM], dpi=cfg.DPI)
        cmap = plt.get_cmap('viridis_r')
        cmap.set_bad('black')

        # plot the rmse values
        sns.barplot(
            data=rmse_df,
            x='Sample ID',
            y='RMSE',
            hue='Grain Size',
            ax=ax, dodge=False)
        plt.xticks(rotation=90)

        # add mineral name labels
        rmse_hi = max(rmse_df.RMSE) * 1.03
        rmse_lo = max(rmse_df.RMSE) * 1.01

        for mineral_name in mineral_names:
            ax.annotate(
                '',
                xy=(mineral_bounds[mineral_name][0]-0.5, rmse_hi),
                xytext=(mineral_bounds[mineral_name][1]+0.5, rmse_hi),
                arrowprops=dict(arrowstyle='<|-|>', shrinkA=0, shrinkB=0),
                annotation_clip=False)
            ax.annotate(
                mineral_name,
                xy=(mineral_label_y[mineral_name], rmse_hi),
                xytext=(mineral_label_y[mineral_name], rmse_hi),
                rotation=45,
                ha='left',
                va='bottom',
                annotation_clip=False)

        # add category labels
        for cat in self.categories:
            ax.annotate(
                '',
                xy=(cat_bounds[cat][0]-0.5, rmse_lo),
                xytext=(cat_bounds[cat][1]+0.5, rmse_lo),
                arrowprops=dict(arrowstyle='<|-|>', shrinkA=0, shrinkB=0),
                annotation_clip=False)
            ax.annotate(
                cat,
                xy=(cat_label_y[cat], rmse_lo),
                xytext=(cat_label_y[cat], rmse_lo),
                rotation=45,
                ha='right',
                va='top',
                annotation_clip=False)

        sns.move_legend(ax, "center left", bbox_to_anchor=(1, 0.5))
        sns.despine()
        plt.tight_layout()

        output_file = Path(self.object_dir / 'analysis',
                            'rmse').with_suffix(cfg.PLT_FRMT)
        plt.savefig(output_file,bbox_inches='tight')

    # """
    # Statistics Overview Functions
    # """

    def statistics(self):
        """Perform statistical analyses over the channels, including,
        correlation, PCA and LDA, and visualise these results.
        """
        if cfg.TIME_IT:
            tic = time.perf_counter()

        print('Performing statistical analyses over all channels...')
        _ = self.channel_correlation()
        _,_ = self.channel_pca()
        _ = self.channel_lda()

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Statistical analyses complete in {toc - tic:0.4f} seconds.")

    def channel_correlation(self, plot_profiles: bool = cfg.PLOT_PROFILES):
        """Compute the correlation matrix between all channels, visualise this,
        and plot a pair-plot / grid-plot of scatterplots for correlations.
        """
        print('Computing channel correlations...')
        corr = self.main_df[self.wvls].corr()
        if plot_profiles:
            self.plot_correlations(corr)
        return corr

    def plot_correlations(self, corr_matrix: np.array):
        """Plot the correlation matrix of channels.

        :param corr_matrix: n_channel x n_channel correlation matrix
        :type corr_matrix: np.array
        """
        if cfg.TIME_IT:
            tic = time.perf_counter()
            print('Plotting channel correlations...')

        if len(self.chnl_lbls) > 20:
            annotate = False
            lw=0
        else:
            annotate = True
            lw=0.5

        # plot the correlation matrix
        Path(self.object_dir / 'statistics' ).mkdir(parents=True, exist_ok=True)
        fig, ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
        ax = sns.heatmap(
                corr_matrix,
                # vmin=-1, vmax=1, center=0,
                # cmap=sns.diverging_palette(20, 220, n=2000),
                cmap='viridis',
                square=True,
                annot=annotate,
                annot_kws={"fontsize": cfg.LABEL_S/2},
                linewidths=lw,
                cbar_kws={'label': 'Pearson Correlation Coeff.', "shrink": 0.8})
        ax.figure.axes[-1].yaxis.label.set_size(cfg.LABEL_S)
        ax.tick_params(axis="x", labelrotation=80,labelright=True)
        ax.set_xlabel('Channel Centre-Wavelengths (nm)', fontsize=cfg.LABEL_S)
        ax.set_ylabel('Channel Centre-Wavelengths (nm)', fontsize=cfg.LABEL_S)
        ax.set_title(f'{len(self.chnl_lbls)} Channel Correlation Matrix',
                     fontsize=cfg.TITLE_S)
        fig.tight_layout()

        output_file = Path(self.object_dir / 'statistics',
                        'channel_correlation_matrix').with_suffix(cfg.PLT_FRMT)
        fig.savefig(output_file,bbox_inches='tight')

        # plot the correlation relationships in a grid plot - takes too long
        # grid_plot = sns.pairplot(
        #                 data=self.main_df,
        #                 vars = self.wvls,
        #                 hue = 'Category',
        #                 height=1.5*cfg.CM,
        #                 aspect=1,
        #                 corner=False,
        #                 plot_kws={"s": 5})
        # grid_plot.fig.suptitle(f'{len(self.chnl_lbls)} Channel Scatterplots')
        # grid_plot.set(xlim=(0,1), ylim=(0,1))
        # output_file = Path(self.object_dir / 'statistics',
        #                               'channel_pairplots').with_suffix('.pdf')
        # grid_plot.savefig(output_file,bbox_inches='tight')
        # if INLINE:
        #     plt.show()
        # else:
        #     plt.close(grid_plot.fig)

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Correlation plots complete in {toc - tic:0.4f} seconds.")

    def channel_pca(self, plot_profiles: bool=cfg.PLOT_PROFILES):
        """Perform Principal Component Analysis on all channels,
        and visualise the variance of the data across the PC's, and the
        distribution of the transformed data in the PCA basis.
        """
        print('Computing and plotting band principal components...')
        data = self.get_refl_df()

        pca = decomposition.PCA() # initiate the PCA decomposition object
        pca.fit(data) # fit the PCA model to the data
        # put PCA transformed data in DataFrame
        cols = list(map(str, np.arange(0, pca.n_components_)+1))
        pca_data_df = pd.DataFrame(
                        data=pca.transform(data),
                        index=self.main_df.index,
                        columns=cols)
        pca_data_df['Category'] = self.main_df['Category'].astype("category")

        if plot_profiles:
            self.plot_channel_pca(pca_data_df)
            self.plot_channel_pca_variance(pca)
            self.plot_channel_pca_weights(pca)

        return pca_data_df, pca

    def plot_channel_pca(self,
            pca_data_df: pd.DataFrame,
            dimension: int=None) -> None:
        """Plot the variance across the PC's, and the distribution of the
        transformed data in the PCA basis.

        :param pca_data_df: PCA transformed data
        :type pca_data_df: pd.DataFrame
        """

        # plot the transformed data on the first 3 Principal Components in 3D
        Path(self.object_dir / 'statistics' ).mkdir(parents=True, exist_ok=True)
        if dimension in set([None, 3]):
            fig = plt.figure(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            ax = fig.add_subplot(projection='3d')
            ax.set_xlabel('PCA 1', fontsize=cfg.LABEL_S)
            ax.set_ylabel('PCA 2', fontsize=cfg.LABEL_S)
            ax.set_zlabel('PCA 3', fontsize=cfg.LABEL_S)
            cat_cols = sns.color_palette()
            i=0
            for cat in self.categories:
                pc_cat_data = pca_data_df.loc[
                    pca_data_df['Category'].values == cat]
                ax.scatter(
                        pc_cat_data['1'].to_numpy(),
                        pc_cat_data['2'].to_numpy(),
                        pc_cat_data['3'].to_numpy(),
                        color=cat_cols[i],
                        label=cat)
                i+=1
            ax.legend()
            plt.title(f'3D PCA Transformation from {len(self.wvls)} Channels',
                      fontsize=cfg.TITLE_S)
            plt.setp(ax.get_legend().get_texts(), fontsize=cfg.LEGEND_S)
            plt.setp(ax.get_legend().get_title(), fontsize=cfg.LEGEND_S)
            fig.tight_layout()

            output_file = Path(self.object_dir /
                        'statistics','channel_pca_3D').with_suffix(cfg.PLT_FRMT)
            fig.savefig(output_file,bbox_inches='tight')

        # plot 2D case of PCA1 vs PCA2
        if dimension in set([None, 2]):
            fig, ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            sns.scatterplot(
                data=pca_data_df,
                x='1',
                y='2',
                hue = 'Category',
                ax=ax)
            ax.set_xlabel('PCA 1', fontsize=cfg.LABEL_S)
            ax.set_ylabel('PCA 2', fontsize=cfg.LABEL_S)
            plt.title(f'2D PCA Transformation from {len(self.wvls)} Channels',
                      fontsize=cfg.TITLE_S)
            plt.setp(ax.get_legend().get_texts(), fontsize=cfg.LEGEND_S)
            plt.setp(ax.get_legend().get_title(), fontsize=cfg.LEGEND_S)
            fig.tight_layout()
            output_file = Path(self.object_dir / 'statistics',
                               'channel_pca_2D').with_suffix(cfg.PLT_FRMT)
            fig.savefig(output_file,bbox_inches='tight')

        # plot the 1D case of PCA1 only
        if dimension in set([None, 1]):
            fig, ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            sns.histplot(data=pca_data_df, x='1',hue='Category', bins=25,ax=ax)
            ax.set_xlabel('PCA 1', fontsize=cfg.LABEL_S)
            ax.set_ylabel('Count', fontsize=cfg.LABEL_S)
            ax.set_title(f'1D PCA Transform from {len(self.wvls)} Channels',
                          fontsize=cfg.TITLE_S)
            plt.setp(ax.get_legend().get_texts(), fontsize=cfg.LEGEND_S)
            plt.setp(ax.get_legend().get_title(), fontsize=cfg.LEGEND_S)
            fig.tight_layout()
            output_file = Path(self.object_dir / 'statistics',
                               'channel_pca_1D').with_suffix(cfg.PLT_FRMT)
            fig.savefig(output_file,bbox_inches='tight')

    def plot_channel_pca_variance(self,
            pca: decomposition.PCA()) -> None:
        """Plot the explained variance ratios

        :param pca: PCA object
        :type pca: decomposition.PCA
        """
        fig, ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
        x_vals = np.arange(0, pca.n_components_)+1
        plt.bar(x_vals, pca.explained_variance_ratio_, log=True)
        plt.xticks(x_vals, x_vals)
        ax.tick_params(axis="x", labelrotation=80,labelright=True)
        ax.set_xlabel('Component', fontsize=cfg.LABEL_S)
        ax.set_ylabel('Explained Variance Ratio (% of Total Variance)',
                      fontsize=cfg.LABEL_S)
        plt.title(f'PCA Explained Variance for {len(self.wvls)} Channels',
                  fontsize=cfg.TITLE_S)
        output_file = Path(self.object_dir, 'statistics',
                           'pca_weights').with_suffix(cfg.PLT_FRMT)
        fig.savefig(output_file,bbox_inches='tight')

    def plot_channel_pca_weights(self,
            pca: decomposition.PCA()) -> None:
        """Plot the 1st PCA component weightings

        :param pca: PCA object
        :type pca: decomposition.PCA
        """
        fig, ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
        sns.barplot(x=self.wvls, y=pca.components_[0], color='tab:blue', ax=ax)
        ax.set_xlabel('Channel Centre Wavelengths (nm)', fontsize=cfg.LABEL_S)
        ax.set_ylabel('Contribution to PCA1', fontsize=cfg.LABEL_S)
        ax.set_title('PCA 1 Projection Coefficients', fontsize=cfg.TITLE_S)
        ax.tick_params(axis="x", labelrotation=80,labelright=True)
        fig.tight_layout()

        output_file = Path(self.object_dir / 'statistics',
                           'pca1_coeffs').with_suffix(cfg.PLT_FRMT)
        fig.savefig(output_file,bbox_inches='tight')

    def channel_lda(self, plot_profiles: bool = cfg.PLOT_PROFILES) -> None:
        """Perform Linear Discriminant Analysis on the channels,
        and visualise the distribution of the transformed data in the LDA basis.
        """
        print('Computing and plotting band linear discriminants...')

        # perform LDA on the band dataset
        lda = LinearDiscriminantAnalysis()
        lda.fit(self.main_df[self.wvls], self.main_df['Category'].values)
        n_classes = len(lda.classes_)
        # put LDA transformed data in a DataFrame
        lda_data_df = pd.DataFrame(
                        data=lda.transform(self.main_df[self.wvls]),
                        index=self.main_df.index,
                        columns=list(map(str, np.arange(0, n_classes-1)+1)))
        lda_data_df['Category'] = self.main_df['Category'].astype("category")

        if plot_profiles:
            self.plot_channel_lda(lda_data_df, lda)

        return lda_data_df

    def plot_channel_lda(self,
            lda_data_df: pd.DataFrame,
            lda: LinearDiscriminantAnalysis) -> None:
        """Plot the distribution of the transformed data in the LDA basis.

        :param lda_data_df: LDA transformed data
        :type lda_data_df: pd.DataFrame
        :param lda: LDA object
        :type lda: LinearDiscriminantAnalysis
        """
        # plot the lDA results
        n_classes = len(lda.classes_)
        n_chnls = len(self.wvls)
        Path(self.object_dir / 'statistics').mkdir(parents=True, exist_ok=True)
        if n_classes == 2:
            # plot the 1st (and only) LDA axis projection
            lda_fig, lda_ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            sns.histplot(
                    data=lda_data_df,
                    x='1',
                    hue=lda_data_df['Category'],
                    bins=25,
                    ax=lda_ax)
            lda_ax.set_xlabel('LDA 1', fontsize=cfg.LABEL_S)
            lda_ax.set_ylabel('Count', fontsize=cfg.LABEL_S)
            lda_ax.set_title(f'LDA Transformation from {n_chnls} Channels',
                             fontsize=cfg.TITLE_S)
            plt.setp(lda_ax.get_legend().get_texts(), fontsize=cfg.LEGEND_S)
            plt.setp(lda_ax.get_legend().get_title(), fontsize=cfg.LEGEND_S)

            lda_fig.tight_layout()

            output_file = Path(self.object_dir / 'statistics',
                               'channel_lda').with_suffix(cfg.PLT_FRMT)
            plt.savefig(output_file,bbox_inches='tight')


            # plot the 1st (and only) LDA band weightings
            lda_w_fig,lda_w_ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            lda_w_ax = sns.barplot(
                            x=self.wvls,
                            y=lda.coef_[0],
                            color='tab:blue',
                            ax=lda_w_ax)
            lda_w_ax.set_xlabel('Channel Centre Wavelength (nm)',
                                 fontsize=cfg.LABEL_S)
            lda_w_ax.set_ylabel('Contribution to LDA1', fontsize=cfg.LABEL_S)
            lda_w_ax.set_title('LDA 1 Projection Coefficients',
                               fontsize=cfg.TITLE_S)
            lda_w_ax.tick_params(axis="x", labelrotation=80,labelright=True)
            lda_w_fig.tight_layout()

            output_file = Path(self.object_dir / 'statistics',
                               'lda1_coeffs').with_suffix(cfg.PLT_FRMT)
            lda_w_fig.savefig(output_file,bbox_inches='tight')
        elif n_classes == 3:
            lda_fig, lda_ax = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            lda_ax = sns.scatterplot(
                    data=lda_data_df,
                    x='1',
                    y='2',
                    hue='Category',
                    ax=lda_ax)
            lda_ax.set_xlabel('LDA 1', fontsize=cfg.LABEL_S)
            lda_ax.set_ylabel('LDA 2', fontsize=cfg.LABEL_S)
            lda_ax.set_title(f'LDA Transformation from {n_chnls} Channels',
                              fontsize=cfg.TITLE_S)
            lda_fig.tight_layout()

            output_file = Path(self.object_dir / 'statistics',
                               'lda_2D').with_suffix(cfg.PLT_FRMT)
            lda_fig.savefig(output_file,bbox_inches='tight')
        elif n_classes == 4:
            # TODO enter code for 3D plot
            print('do something')

    # """
    # Getter/Setter functions
    # """

    def set_refl_data(self, new_refl_data: pd.DataFrame):
        """Set the reflectance data array according to data
        given in an array, with columns labeled by wavelength

        :param new_refl_data: new reflectance data
        :type new_refl_data: pd.DataFrame
        """
        # access the columns given in the new reflectance DataFrame
        new_refl_data.index = self.main_df.index
        self.main_df[self.wvls] = new_refl_data[self.wvls]

    def get_refl_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a copy of the reflectance dataframe subset of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: reflectance data only of the material collection
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        refl_df = subset_df.loc[:, self.wvls[0]:]
        return refl_df

    def get_subset_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a subset of the dataframe, according to selection
        from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: subset of dataframe according to category and mineral
        :rtype: pd.DataFrame
        """
        if (category is not None) and (mineral_name is not None):
            cat_mnrl_selection = (self.main_df.Category == category) & (
                    self.main_df['Mineral Name'] == mineral_name)
            subset_df = self.main_df.loc[cat_mnrl_selection,:]
        elif category is not None:
            cat_selection = self.main_df.Category == category
            subset_df = self.main_df.loc[cat_selection,:]
        elif mineral_name is not None:
            mnrl_selection = self.main_df['Mineral Name'] == mineral_name
            subset_df = self.main_df.loc[mnrl_selection,:]
        else:
            subset_df = self.main_df
        return subset_df

    def get_hdr_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a copy of the header dataframe subset of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: material collection header data
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        hdr_df = subset_df.loc[:, self.material_collection.header_list]
        return hdr_df

    def get_cat_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Returns a copy of the Categories of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: category labels
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        cat_df = pd.DataFrame(subset_df.Category)
        return cat_df

    def get_mineral_list(self,
            category: str=None,
            unique: bool=False) -> List:
        """Get the list of mineral types in the DataFrame, and give the
        list of unique mineral types if requested

        :param category: give mineral list belonging to category only, defaults
            to None
        :type category: str, optional
        :param unique: request unique mineral names only, defaults to False
        :type unique: bool, optional
        :return: all mineral names or unique mineral names in dataframe
        :rtype: List
        """
        if category is not None:
            cat_selection = self.main_df.Category == category
            mineral_list = self.main_df.loc[cat_selection, 'Mineral Name']
        else:
            mineral_list = self.main_df['Mineral Name']
        if unique:
            mineral_list = mineral_list.unique()
        return mineral_list

    def get_chnl_lbl(self, cwl: Union[str, List]) -> Union[str, List]:
        """Get channel label(s) for given centre-wavelength(s)

        :param cwl: centre-wavelength(s) to retrieve channel label(s) for
        :type cwl: Union[str, List]
        :return: channel label(s)
        :rtype: Union[str, List]
        """
        channel_dict = dict(zip(self.wvls, self.chnl_lbls))
        chnl_lbl = channel_dict[cwl]
        return chnl_lbl

    def get_cwl(self, chnl_lbl: List) -> List:
        """Get centre-wavelength(s) for given channel label(s)

        :param cwl: channel label(s) to retrieve for centre-wavelength(s)
        :type cwl: Union[str, List]
        :return: centre-wavelength(s)
        :rtype: Union[float, List]
        """
        cwl_series = pd.Series(data=self.wvls, index=self.chnl_lbls)
        if isinstance(chnl_lbl, str):
            chnl_lbl = [chnl_lbl] # incapsulate in list
        cwl = cwl_series[chnl_lbl].to_numpy()
        return cwl

    # """
    # Export functions
    # """

    def export_main_df(self):
        """Export the dataframe to csv file, and pickle.

        :param pkl_only: Only output a pkl file of the DataFrame
        :type pkl_only: bool, optional
        """
        print('Exporting the Observation Pickle format...')
        table_dir = Path(self.object_dir / 'tables')
        table_dir.mkdir(parents=True, exist_ok=True)

        pkl_file = Path(self.object_dir, 'observation.pkl')
        self.main_df.to_pickle(pkl_file)

        csv_out_file = Path(table_dir, 'observation.csv')
        self.main_df.transpose().to_csv(csv_out_file)

        print('Observation export complete.')

    def plot_profiles(self, categories_only: bool=False, ci: bool=False):
        """Plot the profiles of the materials as sampled by the instrument
        """
        plotter = SpectralLibraryAnalyser(self)
        plotter.plot_profiles(categories_only=categories_only, ci=ci)


"""Instrument and Instrument Builder Classes

Instrument Class:
Hosts the transmission data for the filters of spectral sampling instrument.

Instrument Builder Class:
Produces files listing the instrument sampling central wavelengths, FWHMs,
and channel labels.

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 29-04-2021
"""
import os.path as osp
from pathlib import Path
import glob
from  shutil import rmtree
import time
from typing import Dict, List, Union
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
import sptk.config as cfg

class Instrument():
    """Hosts the transmission data for the channels of spectral sampling
    instrument."""

    def __init__(
            self,
            name: str,
            project_name: str = 'case',
            load_existing: bool=cfg.LOAD_EXISTING,
            plot_profiles: bool = cfg.PLOT_PROFILES,
            export_df: bool = cfg.EXPORT_DF):
        """Constructor for Instrument

        Represents the spectral transmission of the channels of an instrument
        in a DataFrame.

        :param name: the instrument name, to be looked-up from local directory
        :type name: str
        :param project_name: name of project, defaults to 'case'
        :type project_name: str, optional
        :param plot_profiles: plot material profiles, defaults to PLOT_PROFILES
        :type plot_profiles: bool, optional
        :param export_df: export DataFrames, defaults to EXPORT_DF
        :type export_df: bool, optional
        """
        print('Building Instrument...')
        if cfg.TIME_IT:
            tic = time.perf_counter()

        p_dir, p_name = cfg.build_project_directory(project_name, 'instrument')
        self.project_dir = p_dir
        self.project_name = p_name
        self.object_dir = Path(self.project_dir / 'instrument')
        self.wvls = cfg.WVLS
        self.name = name

        if load_existing:
            # check if the instrument has data already in the project directory
            existing_pkl_path = Path(self.object_dir, 'instrument.pkl')
            if osp.isfile(existing_pkl_path):
                print("Loading existing DataFrame for % r..." % name)
                self.main_df = pd.read_pickle(existing_pkl_path)
            else:
                print('No existing DataFrame, building new for % r...' % name)
                self.build_new_instrument(plot_profiles, export_df)
        else:
            print("Building new DataFrame for % r..." % name)
            self.build_new_instrument(plot_profiles, export_df)

        self.filter_ids = self.main_df.index.to_list()

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Instrument built in {toc - tic:0.4f} seconds.")

    def __del__(self, rmdir: bool = False, rmproj: bool = False) -> None:
        """Instrument Destructor method - optionally deletes
        instrument directory and/or entire project directory

        :param rmdir: instruct removal of instrument directory,
            defaults to False
        :type rmdir: bool, optional
        :param rmproj: instruct removal of entire project directory,
            defaults to False
        :type rmproj: bool, optional
        """
        if rmdir:
            name = self.name
            print(f"Deleting {name} Instrument directory...")
            try:
                rmtree(Path(self.project_dir, 'instrument'))
                print(f"{name} Instrument directory deleted.")
            except FileNotFoundError:
                print(f"No {name} Instrument directory to delete.")
        if rmproj:
            name = self.project_name
            print(f"Deleting {name} directory...")
            try:
                rmtree(Path(self.project_dir))
                print(f"{name} directory deleted.")
            except FileNotFoundError:
                print(f"No {name} directory to delete.")

    def build_new_instrument(self,
            plot_profiles: bool = cfg.PLOT_PROFILES,
            export_df: bool = cfg.EXPORT_DF) -> None:
        """Build DataFrame and optionally export and produce plots.

        :param plot_profiles: plot material profiles, defaults to PLOT_PROFILES
        :type plot_profiles: bool, optional
        :param export_df: export DataFrames, defaults to EXPORT_DF
        :type export_df: bool, optional
        """
        inst_data = Instrument.read_instrument_data(self.name)
        self.main_df = Instrument.build_instrument_df(inst_data)
        # optionally produce plots and files of the instrument
        if plot_profiles:
            self.plot_filter_profiles()
        if export_df:
            self.export_main_df()

    @staticmethod
    def read_instrument_data(name: str) -> pd.DataFrame:
        """Find instrument data in the library instrument directory,
        and read into DataFrame

        :return: instrument cwl and fwhm data
        :rtype: pd.DataFrame
        """
        # get allowed instrument names from the list of existing instrument
        # files in the sptk distribution
        search_str = str(cfg.DATA_DIRECTORY / 'instruments' / '*csv')
        inst_files = glob.glob(search_str)
        allowed_names = [osp.splitext(osp.basename(i))[0] for i in inst_files]
        if name not in allowed_names:
            raise ValueError(f'{name} is not in the instrument directory.')
        inst_file = Path(cfg.DATA_DIRECTORY / 'instruments',
                                                name).with_suffix('.csv')
        inst_data = pd.read_csv(inst_file, index_col=0)
        return inst_data

    @staticmethod
    def build_gauss_filter(
            cwl: Union[np.array, float],
            fwhm: Union[np.array,float]) -> np.array:
        """Build Gaussian filter profiles according to the given cwl, fwhm and
        wvls, in parallel.

        :param cwl: Centre wavelength(s) (nm)
        :type cwl: np.array
        :param fwhm: Full-Width at Half Maximum(s) (nm)
        :type fwhm: np.array
        :returns: table of Gaussian transmission profile data
        :rtype: np.array
        """
        sig = fwhm / 2.355482004503 # convert from fwhm to 1-sigma
        # vectorisation: extend cwls, sigs & wvls to match dimensions
        cwls = np.tile(cwl, [cfg.WVLS.shape[0],1])
        sigs = np.tile(sig, [cfg.WVLS.shape[0],1])
        wvls = np.tile(cfg.WVLS, [cwls.shape[1],1]).transpose()
        # compute the Gaussian profiles in parallel
        gauss = np.exp(-np.power(wvls - cwls, 2.) / (2 * np.power(sigs, 2.)))
        return gauss.transpose()

    @staticmethod
    def build_instrument_df(inst_df: pd.DataFrame) -> pd.DataFrame:
        """Builds instrument transmission profiles for filter cwls and fwhms
        using a Gaussian function, and returns in a DataFrame.

        DataFrame format:
         - Columns: filter
         - Rows: wavelengths (according to cfg.SAMPLE_RES)

        :param inst_df: Instrument filter names, cwls and fwhms
        :type inst_df: pd.DataFrame
        :returns: the instrument transmission table
        :rtype: pd.DataFrame
        """
        if 'cwl' in inst_df.columns:
            # build each filter in the instrument in parallel
            cwls = inst_df['cwl'].values
            fwhms = inst_df['fwhm'].values
            out = Instrument.build_gauss_filter(cwls, fwhms)
            # initialise the dataframe according to contents
            init_df =pd.DataFrame(data=out,columns=cfg.WVLS,index=inst_df.index)
        else:
            out = inst_df.to_numpy().T
            # initialise the dataframe according to contents
            init_df = pd.DataFrame(
                            data=out,
                            columns=cfg.WVLS,
                            index=inst_df.columns)
            init_df.index.rename('filter_id')
            inst_df = pd.DataFrame(init_df.idxmax(axis=1), dtype=int)
            inst_df.index.rename('filter_id')
            inst_df = inst_df.rename(columns={0:'cwl'})
            inst_df['fwhm'] = np.NaN
        # concat with cwl and fwhm information (via inst_df)
        main_df = pd.merge(
                    left=inst_df,
                    right=init_df,
                    left_index=True,
                    right_index=True,
                    how='outer')
        return main_df

    def get_trans_df(self) -> pd.DataFrame:
        """Return a copy of the transmission dataframe only

        :return: copy of the master instrument transmission DataFrame
        :rtype: pd.DataFrame
        """
        trans_df = self.main_df.iloc[:, 2:]
        return trans_df.copy()

    def cwls(self) -> pd.Series:
        """Get the centre wavelengths of the instrument channels

        :return: instrument channel centre wavelengths
        :rtype: pd.Series
        """
        return self.main_df.cwl.copy(deep=True)

    def fwhms(self) -> pd.Series:
        """Get the centre wavelengths of the instrument channels

        :return: instrument channel centre wavelengths
        :rtype: pd.Series
        """
        return self.main_df.fwhm.copy(deep=True)

    def get_metrics(self) -> pd.DataFrame:
        """Return a copy of the filter metrics (cwl, fwhm) dataframe only

        :return: copy of the instrument cwl and fwhm DataFrame
        :rtype: pd.DataFrame
        """
        metric_df = self.main_df.iloc[:,:2]
        return metric_df.copy()

    def get_filter_ids(self,
            cwl: Union[float, int, np.array, List]) -> Union[str, List]:
        """Get filter id(s) for given centre-wavelength(s)

        :param cwl: centre-wavelength(s) to retrieve filter id(s) for
        :type cwl: Union[str, List]
        :return: channel label(s)
        :rtype: Union[str, List]
        """
        if isinstance(cwl, (int, float)):
            cwl = [cwl] # incapsulate in list
        filter_series = pd.Series(
                            data=self.cwls().index.to_list(),
                            index=self.cwls().to_list())
        filter_id = filter_series[cwl].to_list()
        return filter_id

    def plot_filter_profiles(self):
        """Plot all filter profiles
        """
        print('Plotting Instrument Transmission...')
        trans_df = pd.melt(self.main_df.reset_index(),
                            id_vars=['cwl', 'fwhm', 'filter_id'])
        fltr_ax_size = (1.3*cfg.FIG_SIZE[0], cfg.FIG_SIZE[1])
        fig, fltr_ax = plt.subplots(figsize=fltr_ax_size, dpi=cfg.DPI)
        plt.rcParams.update({'font.size': 8})
        fltr_ax.set(
            xbound=(cfg.SAMPLE_RES['wvl_min']-10, cfg.SAMPLE_RES['wvl_max']+10),
            ybound=(-0.05,1.15),
            autoscale_on=False)
        sns.lineplot(
            data=trans_df,
            x='variable',
            y='value',
            ax=fltr_ax,
            hue='cwl',
            palette='nipy_spectral',
            linewidth=0.6,
            legend="full")
        fltr_ax.set_xlabel('Wavelength (nm)', fontsize=cfg.LABEL_S)
        fltr_ax.set_ylabel('Transmission', fontsize=cfg.LABEL_S)
        fltr_ax.set_title(
                f'Transmission Profiles ({self.name})', fontsize=cfg.TITLE_S)
        # add minor grid lines at 50 nm intervals and major gridlines
        fltr_ax.get_xaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())
        fltr_ax.grid(True, which='major',axis='both', lw=0.6)
        fltr_ax.grid(True, which='minor',axis='both', lw=0.3)
        label_params = fltr_ax.get_legend_handles_labels()

        if len(self.main_df['cwl']) > 12:
            fltr_ax.get_legend().remove()
        else:
            labels = (self.main_df['cwl'].astype('str') + '±'
                    + (self.main_df['fwhm']/2).astype('str')+' nm').to_list()
            n_ids = len(labels)
            new_label_params = (label_params[0], labels)
            fltr_ax.legend(*new_label_params,
                    loc='center left', bbox_to_anchor=(1, 0.5),
                    ncol=-(-n_ids // 20),
                    fontsize=cfg.LEGEND_S)
        # TODO colour code the filters in a more appropriate way
        for fltr_id in self.main_df.index.to_list():
            fltr_ax.annotate(
                fltr_id,
                (self.main_df.loc[fltr_id,'cwl'], 1.02),
                ha='left',
                annotation_clip=False,
                rotation=60)
        plt.tight_layout()
        output_file = Path(self.object_dir, self.name).with_suffix(cfg.PLT_FRMT)
        fig.savefig(output_file)

        if len(self.main_df['cwl']) > 12:
            # make table of cwl and fwhm
            figl, axl = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            axl.axis(False)
            labels = (self.main_df.index.astype('str') + ', '
                        + self.main_df['cwl'].astype('str') + ' nm, ∆'
                        + self.main_df['fwhm'].astype('str')+' nm').to_list()
            n_ids = len(labels)
            new_label_params = (label_params[0], labels)
            axl.legend(*new_label_params, loc="center",
                    bbox_to_anchor=(0.5, 0.5),
                    ncol=-(-n_ids // 20),
                    fontsize=cfg.LEGEND_S)
            plt.tight_layout()
            output_file = Path(self.object_dir,
                                self.name+'_lgnd').with_suffix(cfg.PLT_FRMT)
            figl.savefig(output_file)

        print('Plots exported to '+str(Path(self.object_dir)))

    def export_main_df(self):
        """Export the Instrument Transmission to CSV and Pickle."""
        if cfg.TIME_IT:
            tic = time.perf_counter()
        print('Exporting the Instrument to CSV and Pickle formats...')
        out_dir = Path(self.object_dir)
        out_dir.mkdir(parents=True, exist_ok=True)
        csv_out_file = out_dir / "instrument.csv"
        self.main_df.transpose().to_csv(csv_out_file)
        pkl_out_file = out_dir / "instrument.pkl"
        self.main_df.to_pickle(pkl_out_file)
        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Instrument exported in {toc - tic:0.4f} seconds.")

class InstrumentBuilder:
    """A Class for constructing a cwl and fwhm look-up table for an instrument
    of a given spectral range and resolution
    """
    def __init__(
            self,
            instrument_name: str,
            instrument_type: str,
            sampling: Union[int, str],
            resolution: float,
            spectral_range: List=None
        ) -> None:

        self.name = instrument_name
        self.instrument_type = instrument_type
        self.sampling = sampling
        self.resolution = resolution
        self.spectral_range = spectral_range

    def build_instrument(self) -> None:
        """Builds the instrument table and exports to file
        """
        if self.instrument_type == 'filter-wheel':
            inst_df = self.generate_filter_band_table()
        elif self.instrument_type == 'aotf':
            inst_df = self.generate_aotf_band_table()
        elif self.instrument_type == 'lvf':
            inst_df = self.generate_lvf_band_table()
        else:
            raise ValueError('Instrument Type not recognised.')
        self.export_instrument(inst_df)

    def generate_filter_band_table(self) -> pd.DataFrame:
        """For a Filter-Wheel type spectrometer,
        given a number of filter positions to occupy and a spectral range,
        generate a list of centre wavelengths distributed across the spectral
        range evenly, and compute FWHM's according to resolution, and format
        these into a table to be read by the Instrument
        class.

        :return: channel cwls and fwhms of instrument
        :rtype: pd.DataFrame
        """
        if self.instrument_type != 'filter-wheel':
            raise ValueError('Instrument type is not filter-wheel. \
            Please generate band table suitable for % r' % self.instrument_type)
        filter_ids = []

        if not isinstance(self.sampling,
                    (list, pd.core.series.Series,np.ndarray)):
            raise ValueError("Sampling type must be array-like giving CWLs of \
                                                            filters in wheel.")

        cwls = np.round(np.array(self.sampling)).astype(int)
        fwhms = cwls.astype(float) / self.resolution

        for idx in range(1,len(cwls)+1):
            filter_id = f'F{idx:02d}'
            filter_ids.append(filter_id)
        inst_df = pd.DataFrame(
                    data={'filter_id':filter_ids,
                                    'cwl': cwls,
                                    'fwhm': fwhms})
        return inst_df

    def generate_lvf_band_table(self) -> pd.DataFrame:
        """For a Linear Variable Filter type spectrometer,
        given a sampling condition and spectral resolving power, generate a list
        of centre wavelengths and full widths at half maximum, and format these
        into a table to be read by the instrument function of the spectral
        parameters toolkit

        :return: channel cwls and fwhms of instrument
        :rtype: pd.DataFrame
        """
        if self.instrument_type != 'lvf':
            raise ValueError('Instrument type is not lvf. \
            Please generate band table suitable for % r' % self.instrument_type)

        filter_ids = []
        cwls = []
        fwhms = []

        wvl_lo = self.spectral_range[0]
        wvl_hi = self.spectral_range[1]
        start_cwl = wvl_lo + (wvl_lo / self.resolution)
        end_cwl = wvl_hi - (wvl_hi / self.resolution)

        if self.sampling == 'nyquist':
            fwhm_si = 0.5
        elif self.sampling == 'critical':
            fwhm_si = 1.0
        elif self.sampling == 'undersampled':
            fwhm_si = 2.0
        elif self.sampling == 'hi-res':
            fwhm_si = 0.3
        else:
            raise ValueError('Sampling criteria not recognised')

        cwl = start_cwl
        i = 1
        while cwl <= end_cwl:
            cwls.append(cwl)
            fwhm = cwl / self.resolution
            fwhms.append(fwhm)
            filter_id = f'S{i:03d}'
            filter_ids.append(filter_id)
            cwl = cwl + (fwhm * fwhm_si)
            i+=1
        inst_df = pd.DataFrame(data={
                                    'filter_id':filter_ids,
                                    'cwl': cwls,
                                    'fwhm': fwhms})
        return inst_df

    def generate_aotf_band_table(self) -> pd.DataFrame:
        """For an Acoust-Optic Tunable Filter type spectrometer,
        given a sampling condition and spectral resolving power gradient with
        wavenumber, generate a list of centre wavelengths and full widths at
        half maximum, and format these into a table to be read by the instrument
        function of the spectral parameters toolkit

        :return: channel cwls and fwhms of instrument
        :rtype: pd.DataFrame
        """
        if self.instrument_type != 'aotf':
            raise ValueError('Instrument type is not aotf. \
            Please generate band table suitable for % r' % self.instrument_type)

        filter_ids = []
        cwls = []
        fwhms = []

        wvl_lo = self.spectral_range[0]
        wvl_hi = self.spectral_range[1]
        res = self.resolution
        start_cwl = wvl_lo + InstrumentBuilder.aotf_cwl_2_fwhm(wvl_lo, res)
        end_cwl = wvl_hi - InstrumentBuilder.aotf_cwl_2_fwhm(wvl_hi, res)

        if self.sampling == 'nyquist':
            fwhm_si = 0.5
        elif self.sampling == 'critical':
            fwhm_si = 1.0
        elif self.sampling == 'undersampled':
            fwhm_si = 2.0
        elif self.sampling == 'hi-res':
            fwhm_si = 0.3
        else:
            raise ValueError('Sampling criteria not recognised')

        cwl = start_cwl
        i = 1
        while cwl <= end_cwl:
            cwls.append(cwl)
            fwhm = InstrumentBuilder.aotf_cwl_2_fwhm(cwl, res)
            fwhms.append(fwhm)
            filter_id = f'S{i:03d}'
            filter_ids.append(filter_id)
            cwl = cwl + (fwhm * fwhm_si)
            i+=1
        inst_df = pd.DataFrame(data={
                        'filter_id':filter_ids,
                        'cwl': cwls,
                        'fwhm': fwhms})
        return inst_df

    def export_instrument(self, inst_df: pd.DataFrame) -> None:
        """Export the instrument table to csv file, and save in sptk directory

        :param inst_df: channel cwls and fwhms of instrument
        :type inst_df: pd.DataFrame
        """
        inst_dir = Path(cfg.DATA_DIRECTORY / 'instruments')
        filepath = Path(inst_dir, f'{self.name}').with_suffix('.csv')
        print(f"Exporting instrument to {filepath}...")
        inst_df.to_csv(filepath, index=False)

    @staticmethod
    def aotf_cwl_2_fwhm(
        cwl: Union[np.float, np.array],
        resolution_model: Dict,
        ) -> Union[np.float, np.array]:
        """Compute the fwhm (nm) given a cwl (nm)

        :param cwl: central wavelength (nm) to compute fwhm at
        :type cwl: Union[n.float, np.array]
        :param resolution_model: linear model for converting wavenumber to
            spectral resolving power
        :type resolution_per_wavenumber: dict
        :return: full-width at half-maximum(maxima), for central wavelength(s)
        :rtype: Union[np.float, np.array]
        """
        cwn = 1E7 / cwl # cm^-1
        resolution_per_wavenumber = resolution_model['m'] # 1/cm^-1
        offset = resolution_model['c']
        power = cwn * resolution_per_wavenumber + offset
        fwhm = cwl / power
        return fwhm



"""Configuration file for Spectral Parameters Toolkit

Sets variables for:
 - timing of procedures
 - sampling resolution

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 18-05-2022
"""
import os
from pathlib import Path
from typing import Tuple
import numpy as np

TIME_IT = False # report process durations

"""Spectral Resolution"""
SAMPLE_RES={"wvl_min": 400, # spectral range (nm)
            "wvl_max": 1100, # spectral range (nm)
            "delta_wvl": 1} # spectral intervals (nm)
WVLS = np.arange(
            SAMPLE_RES['wvl_min'],
            SAMPLE_RES['wvl_max'],
            SAMPLE_RES['delta_wvl'],
            dtype=float)

"""Plotting"""
PLOT_PROFILES = True
# INLINE = False # produce plots inline - i.e. for notebook
CM = 1/2.54  # set scale factor for specifying fig size in cms
FIG_SIZE = (8*CM, 7*CM)
DPI = 300
PLT_FRMT = '.png' # '.png' or '.pdf'
LABEL_S = 8
TITLE_S = 10
LEGEND_S = 6

"""Exports"""
EXPORT_DF = True
LOAD_EXISTING = True # if True load existing directories, else build new
DATA_DIRECTORY = Path('..', 'data')
OUTPUT_DIRECTORY = Path('..', 'projects')

def build_project_directory(
        project_name: str,
        class_name: str) -> Tuple[Path, str]:
    """Builds a directory for the project at the users Desktop.

    If the project already exists, give option for loading in the existing
    project, or writing a new one - specify this with a keyword - default is
    false. Duplicate project names have a number suffix applied to the project
    name.

    :return: project directory path, and the updated project name
    :rtype: Tuple[Path, str]
    """
    project_dir = OUTPUT_DIRECTORY / project_name
    dir_built = False
    pn_w_sffx = project_name
    sffx_n = 2  # original directory is '1', so start suffix at 2

    while not dir_built:
        try: # if the project_dir does not exist, move to exception
            Path(project_dir / class_name).mkdir(parents=True, exist_ok=False)
            project_name = pn_w_sffx
            dir_built = True
        except FileExistsError:
            if LOAD_EXISTING:
                dir_built = True # no new folder will be written
            else: # add suffix number to project name and try again
                pn_w_sffx = project_name + '_' + str(sffx_n)
                project_dir = Path(OUTPUT_DIRECTORY / pn_w_sffx)
                sffx_n += 1

    # if ~LOAD_EXISTING:
    #     Path(project_dir / class_name).mkdir(parents=True, exist_ok=True)

    return project_dir, project_name

def resolve_path(path: Path, root: str='data') -> Path:
    """Resolve a path relative to the data or output directory.

    :param path: Path to resolve
    :type path: Path
    :param root: Root is for data or output directory, defaults to 'data'
    :type root: str, optional
    :return: Complete resolved path
    :rtype: Path
    """
    if root == 'data':
        return DATA_DIRECTORY / path
    elif root == 'output':
        return OUTPUT_DIRECTORY / path
    else:
        raise ValueError('root must be "data" or "output"')


"""Linear Discriminant Analysis Class

Perform key steps of Linear Discriminant Analysis on multiple combinations of
spectral parameter values in parallel.

TODO cythonize this file

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 25-05-2021
"""
from typing import Tuple
import numpy as np
from scipy import linalg
import pandas as pd

def within_class_scatter_matrix(
        k_stack: np.array,
        category_list: pd.Series) -> np.array:
    """Computes the total Within-Category Scatter Matrix S_W, on the dataset
    provided by k_stack, according to category labels.

    Within-category scatter defined as:
        $S_W = Sum(S_c)$
    where S_c is the Within-Class Scatter Matrix of the class 'c', given by:
        $S_c = Sum_{x_in_c}(x - mu_c)^T.(x - mu_c)$
    where x is a vector, given by the dataset, with each element giving the
    value of a feature (or specifically a spectral parameter).
    x_in_c are all of the entries in the dataset that belong to the class 'c',
    and mu_c is the vector that gives the mean feature values in the class 'c'.

    :param k_stack: Data to evaluate. array of spectral parameter
            values of each combination element and each material sample, with
            k_combinations stacked in the 3rd dimension, such that dimensions
            are: [n_combinations x n_samples x k_combinations]
    :type k_stack: np.array
    :param category_list: category label for each sample
    :type category_list: pd.Series
    :return: array of within-class scatter matrices S_w, with spectral parameter
        combination indexed by leading dimension of the array
    :rtype: np.array
    """
    if len(k_stack.shape) != 3:
        k_stack = k_stack.reshape(1,k_stack.shape[0],k_stack.shape[1])
    n_combinations = k_stack.shape[0]
    k_combinations = k_stack.shape[2]

    # initialise within-class scatter matrix (wcsm)
    wcsm = np.zeros((n_combinations, k_combinations, k_combinations))

    # compute wcsm
    for cat in category_list.unique():
        cat_data = k_stack[:, np.where(category_list.to_numpy() == cat)[0], :]
        cat_means = np.mean(cat_data, axis=1, keepdims=True)
        x_centered = cat_data - cat_means
        wcsm += np.matmul((x_centered).transpose(0,2,1), x_centered)
    return wcsm

def between_class_scatter_matrix(
        k_stack: np.array,
        category_list: pd.Series) -> np.array:
    """Computes the total Between-Category Scatter Matrix S_B, on the dataset
    provided by k_stack, according to category labels provided by category_list.

    Between-category scatter matrix defined as:
        $S_B = Sum_{c} n_c.(mu_c - mu).(mu_c - mu)^T$
    where mu_c is the vector that gives the mean feature (spectral parameter)
    values within the class 'c', n_c is the number of entries in the class 'c',
    and mu is the vector that gives the mean feature (spectral parameter) values
     across all classes.

    :param k_stack: Data to evaluate. array of spectral parameter
            values of each combination element and each material sample, with
            k_combinations stacked in the 3rd dimension, such that dimensions
            are: [n_combinations x n_samples x k_combinations]
    :type k_stack: np.array
    :param category_list: category labels for each sample
    :type category_list: pd.Series
    :return: array of between-class scatter matrices S_b, with spectral
        parameter combination indexed by leading dimension of the array
    :rtype: np.array
    """
    # ensure the leading dimension indexes the spectral parameter combinations
    if len(k_stack.shape) != 3:
        k_stack = k_stack.reshape(1,k_stack.shape[0],k_stack.shape[1])
    n_combinations = k_stack.shape[0]
    k_combinations = k_stack.shape[2]

    # initialise between-class scatter matrix (bcsm)
    bcsm = np.zeros((n_combinations, k_combinations, k_combinations))

    # compute bcsm
    for cat in category_list.unique():
        cat_data = k_stack[:, np.where(category_list.to_numpy() == cat)[0], :]
        cat_means = np.mean(cat_data, axis=1, keepdims=True)
        grand_means = np.mean(k_stack, axis=1, keepdims=True)
        n_c = len(np.where(category_list.to_numpy() == cat)[0])
        cntrd_means = cat_means - grand_means
        bcsm += n_c*np.matmul(cntrd_means.transpose(0,2,1), cntrd_means)

    return bcsm

def projection_matrix(
        wcsm: np.array,
        bcsm: np.array,
        n_classes: int = 2,
        singular_spcs: np.array=None) -> Tuple[np.array, np.array, np.array]:
    """Compute the projection matrix A as the solution to the problem
        $S_B a_i = lambda_k S_W a_i$
    where a_i is the projection vector for each LDA axis, of which there are
    n_c - 1 axes, where n_c is the number of classes.

    This is solved by finding the eigenvalues and eigenvectors of $S_W^-1.S_B$,
    where S_W is the within-class scatter and S_B the between-class scatter.

    See Duda, Hart & Stork, Pattern Classification, §4.11, pp. 47 - 51

    S_W and S_B are [..., k_combinations, k_combinations] matrices, where the
    [...] dimension denotes that the operation can act on multiple sets of data
    in parallel.
    The returned projection matrix has dimensions:
    [..., k_combinations, n_classes - 1]

    :param wcsm: array of within-class scatter matrices S_w, with spectral
        parameter combination indexed by leading dimension of the array
    :type wcsm: np.array
    :param bcsm: array of between-class scatter matrices S_w, with spectral
        parameter combination indexed by leading dimension of the array
    :type bcsm: np.array
    :param n_classes: number of categories to classify, defaults to 2
    :type n_classes: int, optional
    :return: array of LDA solutions for projection coordinates,
        and eigenvalues and eigenvectors used to compute these,
        with spectral parameter combination indexed by leading dimension of
        the array.
    :rtype: Tuple[np.array, np.array, np.array]
    """
    # ensure the leading dimensions index the spectral parameter combinations
    if len(wcsm.shape) != 3:
        k_combinations = wcsm.shape[0]
        wcsm = wcsm.reshape(1,k_combinations,k_combinations)
    if len(bcsm.shape) != 3:
        k_combinations = bcsm.shape[0]
        bcsm = bcsm.reshape(1,k_combinations,k_combinations)

    # initialise the output projection matrix
    n_combinations = wcsm.shape[0]
    k_combinations = wcsm.shape[2]
    projection = np.zeros((n_combinations, k_combinations, n_classes-1))

    singular_mask = singular_spcs # new way of filtering spcs
    if singular_mask is None:
        singular_mask = np.zeros(n_combinations, dtype=bool)
    projection[singular_mask,0,:] = 1.0

    # solve eigen-problem
    nonsing_inv_s_w = np.linalg.inv(wcsm[~singular_mask])
    nonsing_s_b = bcsm[~singular_mask]
    eig_vals, eig_vecs = np.linalg.eig(np.matmul(nonsing_inv_s_w, nonsing_s_b))
    eig_vals = eig_vals.real # get real values only
    eig_vecs = eig_vecs.real # get real values only

    # sort the eigenvectors by size of eigenvalue
    idx_srt = np.argsort(-eig_vals, axis=1)
    idx_vec_srt = np.repeat(idx_srt[:, np.newaxis, :], k_combinations, axis=1)
    eig_vals = np.take_along_axis(eig_vals, idx_srt, axis=1)
    eig_vecs = np.take_along_axis(eig_vecs, idx_vec_srt, axis=2)

    # Get the n_classes - 1 largest eigenvalue associated eigenvectors and
    # put in projection matrix
    projection[~singular_mask] = eig_vecs[:,:,0:n_classes-1].copy()

    return projection, eig_vecs, eig_vals

def fisher_ratio(
        projections: np.array,
        bcsm: np.array,
        wcsm: np.array) -> np.array:
    """Computes the Fisher Ratio:

        $J(A) = |A^T S_B A| / |A^T S_W A|$

    where A is the [..., n_features, n_classes - 1] projection matrix,
    S_B is the [..., n_features, n_features] between-class scatter matrix,
    S_W is the [..., n_features, n_features] within-class scatter matrix,
    and '...' denotes that the operation can be performed on multiple sets
    of data in parallel, provided n_features and n_classes are the same for
    each.
    Returns Fisher-Ratio as a [..., 1] vector.

    When n_classes > 2, J(A) ('score') is an [n_classes-1, n_classes-1] matrix.

    :param projections: array of LDA solutions for projection coordinates, with
        spectral parameter combination indexed by leading dimension of the array
    :type projections: np.array
    :param bcsm: array of between-class scatter matrices S_w, with spectral
        parameter combination indexed by leading dimension of the array
    :type bcsm: np.array
    :param wcsm: array of within-class scatter matrices S_w, with spectral
        parameter combination indexed by leading dimension of the array
    :type wcsm: np.array
    :return: Fisher-Ratio score for each spectral parameter combination computed
         on the dataset
    :rtype: np.array
    """
    between_class_scatter = np.matmul(
                                projections.transpose(0,2,1),
                                np.matmul(bcsm, projections))
    within_class_scatter = np.matmul(
                                projections.transpose(0,2,1),
                                np.matmul(wcsm, projections))
    score = np.divide(np.linalg.det(between_class_scatter),
                                        np.linalg.det(within_class_scatter))
    return score

def project_data(
        k_stack: np.array,
        projections: np.array) -> np.array:
    """Project the data onto the LDA axis for each spectral parameter
    combination.

    :param k_stack: Data to evaluate. array of spectral parameter
            values of each combination element and each material sample, with
            k_combinations stacked in the 3rd dimension, such that dimensions
            are: [n_combinations x n_samples x k_combinations]
    :type k_stack: np.array
    :param projections: array of LDA solutions for projection coordinates, with
        spectral parameter combination indexed by leading dimension of the array
    :type projections: np.array
    :return: data for each sample projected on the LDA axis for each spectral
        parameter combination
    :rtype: np.array
    """
    projected_data = np.sum(k_stack * projections[:, np.newaxis,:], axis=2).T
    return projected_data

def compute_lda_boundary(
        projection,
        k_stack: np.array,
        category_list: pd.Series
        ) -> Tuple[pd.Series, pd.Series]:
    """Compute the decision boundary for each spectral parameter combination
        LDA projection, and determine if the target category is greater than
        or less than the background category on the LDA axis.

    :param projections: array of LDA solutions for projection coordinates, with
        spectral parameter combination indexed by leading dimension of the array
    :type projections: np.array
    :param k_stack: 3D array of spectral parameter values, with
        k_combinations stacked in the 3rd dimension, such that dimensions
        are: [n_combinations x n_samples x k_combinations]
    :type k_stack: np.array
    :param category_list: category label for each sample
    :type category_list: pd.Series
    :return: boundary value, and boolean of target>background, for each
        spectral parameter combination.
    :rtype: Tuple[pd.Series, pd.Series]
    """
    projected_data = project_data(k_stack, projection)
    projected_df = pd.DataFrame(data=projected_data, index=category_list)
    cat_means = projected_df.groupby('Category').mean()

    # very simple estimate of boundary for 2 class case, assuming equal variance
    boundaries = cat_means.sum() / 2

    background_mean = cat_means.loc[category_list.cat.categories[0]]
    target_mean = cat_means.loc[category_list.cat.categories[1]]

    target_gt_background_bools = target_mean > background_mean

    return boundaries, target_gt_background_bools


"""Material Collection Class

Hosts material reflectance data and auxilary information in a Pandas DataFrame

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 29-04-2021
"""

import errno
import os
import glob
import time
from pathlib import Path
from  shutil import rmtree
from typing import Dict, List
import click
import pandas as pd
import numpy as np
from scipy.interpolate import interp1d
import sptk.config as cfg
from sptk.instrument import Instrument
from sptk.spectral_library_analyser import SpectralLibraryAnalyser

HEADER_LIST = [
    'Sample ID', 'Mineral Name',
    'Sample Description', 'Date Added', 'Viewing Geometry',
    'Other Information', 'Formula', 'Composition',
    'Resolution', 'Grain Size', 'Locality',
    'Database of Origin']

class MaterialCollection():
    """Hosts material reflectance data and auxilary information in a DataFrame.
    """

    def __init__(
            self,
            materials: dict,
            spectral_library: str = None,
            project_name: str = 'case',
            load_existing: bool = cfg.LOAD_EXISTING,
            balance_classes: bool=True,
            random_bias_seed: int = None,
            allow_out_of_bounds: bool=False,
            plot_profiles: bool = cfg.PLOT_PROFILES,
            export_df: bool = cfg.EXPORT_DF) -> None:
        """Constructor for MaterialCollection class

        :param materials: dictionary mapping of class labels to mineral group
            names and filenames or wildcards.
            e.g.: materials = {
                        'class_1': [('material_1', file_specification)],
                        'class_2': [('material_2', file_specification),
                                    ('material_3', file_specification)]
                                }
        :type materials: dict
        :param spectral_library: name of spectral library to draw samples from,
            defaults to None
        :type spectral_library: str, optional
        :param project_name: name of project and directory, defaults to 'case'
        :type project_name: str, optional
        :param load_existing: instruct to use or overwrite existing directories
            and files of the same project_name, defaults to config.py setting.
        :type load_existing: bool, optional
        :param balance_classes: instruct whether to balance class sizes by
            randomly removing samples until class sizes are equal, defaults to
            True.
        :type balance_classes: bool, optional
        :param random_bias_seed: only valid if balance_classes is True; set to
            an integer for reproducible balancing, set to None for random
            balancing, defaults to None.
        :type random_bias_seed: int, optional
        :param allow_out_of_bounds: Instruct whether to include entries with
            wavelengths that do not cover the full range, defaults to False.
        :type allow_out_of_bounds: bool, optional
        :param plot_profiles: plot material profiles to project directory,
            defaults to False
        :type plot_profiles: bool, optional
        :param export_df: export DataFrames to project directory, defaults to
            False
        :type export_df: bool, optional
        """
        if cfg.TIME_IT:
            tic = time.perf_counter()

        self.project_dir, self.project_name = cfg.build_project_directory(
            project_name, 'material_collection')
        self.object_dir = Path(self.project_dir / 'material_collection')
        self.material_file_dict = MaterialCollection.parse_materials(
            materials, spectral_library)
        self.categories = list(self.material_file_dict.keys())
        self.wvls = cfg.WVLS
        self.allow_out_of_bounds = allow_out_of_bounds
        self.header_list = HEADER_LIST

        if load_existing:
            existing_pkl_path = Path(
                    self.object_dir,
                    'material_collection').with_suffix('.pkl')
            file_exists = os.path.isfile(existing_pkl_path)
            if file_exists:
                print(f"Loading existing {project_name} MaterialCollection DF")
                self.main_df = pd.read_pickle(existing_pkl_path)
            else:
                print(f"No existing {project_name} MaterialCollection DF")
                print("Building new MaterialCollection")
                self.main_df = self.build_new_material_collection()
        else:
            print(f"Building new {project_name} MaterialCollection DF")
            self.main_df = self.build_new_material_collection()

        if balance_classes:
            self.balance_class_sizes(random_state = random_bias_seed)
        if plot_profiles:
            plotter = SpectralLibraryAnalyser(self)
            plotter.plot_profiles()
        if export_df:
            self.export_main_df()

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Material Collection built in {toc - tic:0.4f} seconds.")

    def __del__(self, rmdir: bool = False, rmproj: bool = False) -> None:
        """MaterialCollection Destructor object - optionally deletes
        material_collection directory and/or entire project directory

        :param rmdir: instruct removal of material_collection directory,
            defaults to False
        :type rmdir: bool, optional
        :param rmproj: instruct removal of entire project directory,
            defaults to False
        :type rmproj: bool, optional
        """
        name = self.project_name
        if rmdir:
            print(f"Deleting {name} MaterialCollection directory...")
            try:
                rmtree(Path(self.project_dir, 'material_collection'))
                print(f"{name} MaterialCollection directory deleted.")
            except FileNotFoundError:
                print(f"No {name} MaterialCollection directory to delete.")
        if rmproj:
            print(f"Deleting {name} directory...")
            try:
                rmtree(Path(self.project_dir))
                print(f"{name} directory deleted.")
            except FileNotFoundError:
                print(f"No {name} directory to delete.")

    @staticmethod
    def parse_materials(material_dict: dict, spectral_library: str) -> dict:
        """Parses the 'materials' dictionary (category labels for material names
        and entry IDs) into a dictionary of class labels and entry filepaths,
        for look-up in the specified local spectral library.

        :param materials: dictionary mapping of class labels to mineral group
            names and filenames or wildcards.
            e.g.: materials = {
                        'class_1': [('material_1', file_specification)],
                        'class_2': [('material_2', file_specification),
                                    ('material_3', file_specification)]
                                }
        :type materials: dict
        :param spectral_library: name of spectral library to draw samples from,
            defaults to None
        :type spectral_library: str, optional
        :return: Dictionary mapping class labels to lists of entry filepaths
        :rtype: dict
        """
        data_dir = cfg.DATA_DIRECTORY
        if spectral_library is None:
            data_library = Path(data_dir, 'spectral_library')
        else:
            data_library = Path(data_dir, 'spectral_library', spectral_library)

        data_library_exists = os.path.isdir(cfg.resolve_path(data_library, root='data'))
        if not data_library_exists:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), str(data_library))

        entry_dict = {}
        for cat in material_dict.keys():
            # for each category make new dict entry
            cat_path = data_library

            entries = material_dict[cat]
            entry_is_singular = isinstance(entries, tuple)
            if entry_is_singular:
                entries = [entries] # encase in list

            cat_entries = []
            for entry in entries:  # build entry filepath
                mineral_group = entry[0]
                file_keys = entry[1]
                file_keys_in_list = isinstance(file_keys, list)
                if not file_keys_in_list:
                    file_keys = [file_keys] # encase in list
                entry_files = []
                for file_key in file_keys:  # add category and material to list
                    filename = file_key + '.csv'
                    entry_files.append(cat_path / mineral_group / filename)
                # expand the filenames into list
                # cat_entries = cat_entries + \
                #     sum([glob.glob(str(cfg.resolve_path(entry_file, root='package')))
                #             for entry_file in entry_files], [])
                # expand the filenames into list, but relative to the package
                for entry_file in entry_files:
                    file_list = glob.glob(str(cfg.resolve_path(entry_file, root='data')))
                    file_list = [os.path.relpath(file, cfg.DATA_DIRECTORY) for file in file_list]
                    cat_entries = cat_entries + file_list

            entry_dict[cat] = sorted(cat_entries)

        return entry_dict

    def build_new_material_collection(self) -> pd.DataFrame:
        """Initialises and populates a new MaterialCollection DataFrame, and
        optionally balances the classes, produces plots, and exports to csv.

        Material files with out-of-range wavelengths will have values set to
        None in the DataFrame.
        Spectral reflectance is interpolated to wavelength resolution set in
        config.py.
        """
        main_df = MaterialCollection.init_frame(
            self.categories, self.material_file_dict)
        filepaths = main_df['Filepath'].tolist()
        with click.progressbar(filepaths) as load_bar:
            for filepath in load_bar:  # for each file, load_material
                new_entry = MaterialCollection.load_material(filepath)
                filename = os.path.basename(filepath)
                mat_group = os.path.basename(os.path.dirname(filepath))
                if not self.allow_out_of_bounds:
                    has_nan = np.isnan(np.sum(new_entry[cfg.WVLS].to_numpy()))
                    if has_nan:
                        print(f'{mat_group}: {filename} is out of bounds, removing...')
                        main_df.drop(
                            main_df[main_df.Filepath == filepath].index,
                            inplace=True)
                    else:
                        print(f'{mat_group}: {filename} loaded')
                        main_df.loc[main_df.Filepath == filepath,
                                        new_entry.columns] = new_entry.values
                else:
                    print(f'{mat_group}: {filename} loaded')
                    main_df.loc[main_df.Filepath == filepath,
                                    new_entry.columns] = new_entry.values
        # put the Data ID as the index
        main_df.set_index('Data ID', inplace=True)
        print('Loading of entries complete.')
        return main_df

    @staticmethod
    def init_frame(
            categories: List,
            material_file_dict: Dict,
            ) -> pd.DataFrame:
        """Initialise the Material Collection DataFrame with the list of entry
        filepaths and associated Categories, with pre-defined feature labels.

        DataFrame format:
         - Columns: material properties and reflectance wavelengths
         - Rows: material samples

        :param categories: categories of the material collection
        :type categorires: List
        :param material_file_dict: dictionary mapping categories to materials
            and entry filepaths in the spectral library
        :type material_file_dict: Dict
        :return: Initiliased Material Collection DataFrame.
        :rtype: pd.DataFrame
        """
        filepaths = []
        material_categories = []
        # build lists of material files, and associated categories
        for category in categories:
            cat_filepaths = material_file_dict[category]
            for material_filepath in cat_filepaths:
                filepaths.append(material_filepath)
                material_categories.append(category)
        # build null list as placeholder for feature column list
        null_list = [''] * len(filepaths)
        empt_df = pd.DataFrame({  # put lists in dataframe
            'Filepath': pd.Series(filepaths, dtype='str'),
            'Category': pd.Series(material_categories, dtype='category'),
            'Data ID': pd.Series(null_list, dtype='str'),
            'Sample ID': pd.Series(null_list, dtype='str'),
            'Mineral Name': pd.Series(null_list, dtype='str'),
            'Sample Description': pd.Series(null_list, dtype='str'),
            'Date Added': pd.Series(null_list, dtype='str'),
            'Viewing Geometry': pd.Series(null_list, dtype='str'),
            'Other Information': pd.Series(null_list, dtype='str'),
            'Formula': pd.Series(null_list, dtype='str'),
            'Composition': pd.Series(null_list, dtype='str'),
            'Resolution': pd.Series(null_list, dtype='str'),
            'Grain Size': pd.Series(null_list, dtype='str'),
            'Locality': pd.Series(null_list, dtype='str'),
            'Database of Origin': pd.Series(null_list, dtype='str'),
            'Reflectance': pd.Series(['---']*len(filepaths), dtype='str')})
        n_entries = sum([len(files) for files in material_file_dict.values()])
        refl_idx = pd.RangeIndex(stop=n_entries)
        refl_df = pd.DataFrame(
                np.nan, index=refl_idx, columns=cfg.WVLS)  # reflectance DF
        empt_df = pd.concat([empt_df, refl_df], axis=1)
        return empt_df

    @staticmethod
    def load_material(filepath: str) -> pd.DataFrame:
        """Access material file in the sptk spectral library, and parse data
        into DataFrame suitable for appending to master dataframe.

        notation: 'mtrl' is 'material'

        :param material_filepath: filepath location of the material
        :type material_filepath: str
        :return: material reflectance data and metadata
        :rtype: pd.DataFrame
        """
        resolved_filepath = cfg.resolve_path(filepath, root='data')
        mtrl_in = pd.read_csv(resolved_filepath,
                              header=None,
                              index_col=0,
                              engine='c',
                              on_bad_lines='skip')

        # resample the wavelengths to the specified range
        mtrl_refl_in = mtrl_in.loc['Wavelength':]
        mtrl_refl_in = mtrl_refl_in.drop(mtrl_refl_in.index[0])
        wvls_in = mtrl_refl_in.index.to_numpy().astype('float64')

        # perform interpolation; put NaN in bad values
        refl_in = mtrl_refl_in.to_numpy().astype('float64')
        refl_in = np.reshape(refl_in, len(refl_in))
        refl_func = interp1d(wvls_in, refl_in, bounds_error=False)
        refl_interp = refl_func(cfg.WVLS)
        refl_series = pd.Series(data=refl_interp, index=cfg.WVLS)

        # parse the material header information
        hdr = mtrl_in.loc[:'Wavelength']

        # replace 'Sample Name' with 'Mineral Name'
        hdr = hdr.rename({'Sample Name': 'Mineral Name'})

        # replace the Mineral Name with the directory name
        mnrl_dir_name = os.path.split(os.path.split(filepath)[0])[1]
        hdr.loc['Mineral Name'] = mnrl_dir_name

        try:
            # if no 'Data ID' label, try 'Sample ID'
            hdr.loc['Data ID']
        except KeyError:
            # use filename as Data ID
            hdr.loc['Data ID'] = Path(filepath).stem

        # handling of new Grain Size and Grain Size Description entries to VISOR
        # discards Grain Size in favour of 'old' style Grain Size Description,
        # but keeps label of 'Grain Size'.
        # TODO: Problem - new ViSOR file formats use '(X, Y)' to define range of
        # grain sizes, so we have a comma in a field, causing the Pandas csv
        # parser to fail. Current hack fix is to just skip the bad lines, so
        # that the reflectance data still loads correctly.
        # To fix properly, we need a way to define the information within the
        # brackets as a single field.
        # If the parser fails, need to try to wrap the '(X, Y)' grain
        # information in "" quotes, and read again using quotchar = "". Where
        # the entries for the file have '(...)', convert to '"(...)"' and retry.

        props = hdr.index
        if 'Grain Size' in props and 'Grain Size Description' in props:
            hdr.loc['Grain Size'] = hdr.loc['Grain Size Description']
        elif 'Grain Size' not in props and 'Grain Size Description' in props:
            hdr.loc['Grain Size'] = hdr.loc['Grain Size Description']

        # get header data that matches specified HEADER_LIST
        header_indx = pd.Index(['Data ID'] + HEADER_LIST)
        mtrl_hdr = hdr[hdr.index.isin(header_indx)].squeeze()
        # add HEADER_LIST features that are not in the imported headings list
        missing_header = pd.Series(
            data=[None]*len(header_indx[~header_indx.isin(mtrl_hdr.index)]),
            index=header_indx[~header_indx.isin(mtrl_hdr.index)],
            dtype=np.float64)
        mtrl_hdr = pd.concat([mtrl_hdr, missing_header])
        mtrl_hdr = mtrl_hdr.reindex(header_indx)  # apply the original headings
        mtrl_hdr = mtrl_hdr.str.strip() # remove leading/trailing white space
        mtrl_hdr['Mineral Name'] = mtrl_hdr['Mineral Name'].lower()

        # prepare data for appending to main dataframe
        mtrl_df = pd.concat([mtrl_hdr, refl_series]).to_frame().transpose()

        return mtrl_df

    def balance_class_sizes(self, random_state: int = None):
        """Checks for class balance, and randomly removes samples so that all
        classes are of the same size, matching that of the smallest class.

        If a Bayesian method of calculating the decision boundary were
        implemented, then this step wouldn't be necessary.

        :param random_state: set random number seed for reproducibility,
            defaults to None
        :type random_state: int, optional
        """
        if self.main_df.Category.value_counts().is_unique:
            print('Balancing class sizes...')
            class_n_dict = self.main_df.Category.value_counts().to_dict()
            # for each class that is not the smallest class
            min_class = min(class_n_dict, key=class_n_dict.get)
            for cat in self.categories:
                if cat is min_class:
                    continue
                # number of samples to remove
                n_r = class_n_dict[cat] - class_n_dict[min_class]
                # randomly select n_r samples
                to_drop = self.main_df[self.main_df.Category == cat].sample(
                    n=n_r, random_state=random_state).index
                # remove the selected samples
                self.main_df.drop(to_drop, inplace=True)

    def balance_mineral_groups(self, random_state: int=None):
        """Checks for class and mineral group balance. First balances mineral
        groups, then balances classes by stratified undersampling of the
        mineral groups of the larger class size.

        :param random_state: set random number seed for reproducibility,
            defaults to None
        :type random_state: int, optional
        """
        # for each class
        for cat in self.categories:
            cat_df = self.main_df[self.main_df.Category == cat]
            mnrl_n = cat_df.groupby('Mineral Name')['Mineral Name'].count()
            mnrl_r = mnrl_n - mnrl_n.min() # n entries to remove
            for mnrl in mnrl_r.index:
                to_drop = cat_df[cat_df['Mineral Name'] == mnrl].sample(
                                n=mnrl_r[mnrl], random_state=random_state).index
                # remove the selected samples
                self.main_df.drop(to_drop, inplace=True)
        # check difference between class sizes
        if self.main_df.Category.value_counts().is_unique:
            print('Balancing class sizes...')
            class_n_dict = self.main_df.Category.value_counts().to_dict()
            # get the smallest class
            min_class = min(class_n_dict, key=class_n_dict.get)
            # for each class that is not the smallest class:
            for cat in self.categories:
                if cat is min_class:
                    continue
                # number of samples to remove from this class is the
                # difference between the number of samples this class and the
                # number of samples in the smallest class
                n_r = class_n_dict[cat] - class_n_dict[min_class]

                # get the number of samples in each mineral group of this class
                cat_df = self.main_df[self.main_df.Category == cat]
                mnrl_n = cat_df.groupby('Mineral Name')['Mineral Name'].count()

                # number of samples to remove from each mineral group
                n_m_r = n_r // len(mnrl_n)
                # remainder of samples to remove from the class
                n_r_r = n_r % len(mnrl_n)

                # remove number of samples from each mineral group
                for mnrl in mnrl_n.index:
                    # randomly select n_m_r samples
                    to_drop = cat_df[cat_df['Mineral Name'] == mnrl].sample(
                        n=n_m_r, random_state=random_state).index
                    # remove the selected samples
                    self.main_df.drop(to_drop, inplace=True)
                    # update the cat_df to reflect the removal
                    cat_df = self.main_df[self.main_df.Category == cat]
                # remove remainder from this class
                if n_r_r != 0:
                    to_drop = cat_df.sample(
                            n=n_r_r, random_state=random_state).index
                        # remove the selected samples
                    self.main_df.drop(to_drop, inplace=True)

    def print_category_entries(self, print_minerals: bool=False) -> None:
        """Print the number of entries in each category, and optionally
        print the number of entries in each mineral group.

        :param print_minerals: print no. of entries in mineral groups,
            defaults to False
        :type print_minerals: bool, optional
        """
        for cat in self.categories:
            cats = self.main_df[self.main_df.Category == cat]['Sample ID']
            print(f"Samples in {cat} class: {len(cats)}")
            print(f"Unique sample sources in {cat} class: {len(cats.unique())}")
            if print_minerals:
                mnrls = self.main_df[self.main_df.Category==cat]['Mineral Name']
                for mnrl in mnrls.unique():
                    n_mnrls = len(mnrls[mnrls == mnrl])
                    print(f"  Samples in {mnrl} mineral group: {n_mnrls}")

    def sample(self, instrument: Instrument) -> np.array:
        """Sample the material collection with a given instrument.

        :param instrument: the instrument used to perform spectral sampling
        :type instrument: Instrument object
        :return: matrix of sample results
        :rtype: np.array
        """
        print('Sampling the Material Collection with the Instrument...')
        if cfg.TIME_IT:
            tic = time.perf_counter()
        # prepare material collection for sampling
        mat_refl = self.get_refl_df().to_numpy()
        if self.allow_out_of_bounds:
            # log locations of NaN entries
            channel_mask = self.channel_mask(mat_refl, instrument)
            mat_refl = np.nan_to_num(mat_refl)
        # prepare instrument transmission for performing the sampling
        cam_trans = instrument.get_trans_df().T.to_numpy()
        # compute matrix mult. of reflectance against channel transmission
        product = np.matmul(mat_refl, cam_trans)
        # compute sum of filter transmissions over wavelength dimension
        trans_sum = cam_trans.sum(axis=0)
        # reformat sum to match matrix product dimensions
        trans_sum = np.tile(trans_sum, [product.shape[0], 1])
        # normalise matrix multiplication product by filter summation
        sample_matrix = np.divide(product, trans_sum)
        if self.allow_out_of_bounds:
            # apply NaN mask
            sample_matrix[~channel_mask] = np.NaN
        print('Sampling complete.')
        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Observation sampled in {toc - tic:0.4f} seconds.")
        return sample_matrix

    def channel_mask(self,
            mat_refl: np.array,
            instrument: Instrument) -> np.array:
        """Produces an array of masks, labelling channels with NaN entries,
        so that these channels can be neglected when sampling with the
        instrument in the case where allow_out_of_bounds is true.

        :param mat_refl: material collection reflectance array
        :type mat_refl: np.array
        :param instrument: Instrument object
        :type instrument: Instrument
        :return: Boolean array of channels to include for each entry
        :rtype: np.array
        """
        min_wvl = []
        max_wvl = []
        # for i in range(0,len(mat_refl)):
        #     idx = np.where(~np.isnan(mat_refl[i]))[0]
        #     try:
        #         min_nan = idx[0]
        #         min_wvl.append(self.wvls[min_nan])
        #     except IndexError:
        #         min_wvl.append(self.wvls[0])
        #     try:
        #         max_nan = idx[-1]
        #         max_wvl.append(self.wvls[max_nan])
        #     except IndexError:
        #         max_wvl.append(self.wvls[0])

        for i, _ in enumerate(mat_refl):
            idx = np.where(~np.isnan(mat_refl[i]))[0]
            try:
                min_nan = idx[0]
                min_wvl.append(self.wvls[min_nan])
            except IndexError:
                min_wvl.append(self.wvls[0])
            try:
                max_nan = idx[-1]
                max_wvl.append(self.wvls[max_nan])
            except IndexError:
                max_wvl.append(self.wvls[0])

        min_chnls = (instrument.cwls() - instrument.fwhms()).to_numpy()
        max_chnls = (instrument.cwls() + instrument.fwhms()).to_numpy()

        # broadcast min_chnls, max_chnls, min_wvl & max_wvl to match dimensions
        min_wvl = np.tile(min_wvl, (len(min_chnls), 1)).T
        max_wvl = np.tile(max_wvl, (len(max_chnls), 1)).T
        min_chnls = np.tile(min_chnls, (len(mat_refl), 1))
        max_chnls = np.tile(max_chnls, (len(mat_refl), 1))

        min_mask = np.less(min_wvl, min_chnls)
        max_mask = np.greater(max_wvl, max_chnls)

        channel_mask = min_mask * max_mask

        return channel_mask

    def get_main_df(self) -> pd.DataFrame:
        """Return a copy of the main dataframe.

        :return: copy of the main material collection DataFrame
        :rtype: pd.DataFrame
        """
        return self.main_df.copy()

    def set_refl_data(self, new_refl_data: pd.DataFrame):
        """Set the reflectance data array according to data
        given in an array, with columns labeled by wavelength

        :param new_refl_data: new reflectance data
        :type new_refl_data: pd.DataFrame
        """
        # access the columns given in the new reflectance DataFrame
        new_refl_data.index = self.main_df.index
        self.main_df[self.wvls] = new_refl_data[self.wvls]

    def get_refl_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a copy of the reflectance dataframe subset of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: reflectance data only of the material collection
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        refl_df = subset_df.loc[:, cfg.SAMPLE_RES['wvl_min']:]
        return refl_df

    def get_subset_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a subset of the dataframe, according to selection
        from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: subset of dataframe according to category and mineral
        :rtype: pd.DataFrame
        """
        if (category is not None) and (mineral_name is not None):
            cat_mnrl_selection = (self.main_df.Category == category) & (
                    self.main_df['Mineral Name'] == mineral_name)
            subset_df = self.main_df.loc[cat_mnrl_selection,:]
        elif category is not None:
            cat_selection = self.main_df.Category == category
            subset_df = self.main_df.loc[cat_selection,:]
        elif mineral_name is not None:
            mnrl_selection = self.main_df['Mineral Name'] == mineral_name
            subset_df = self.main_df.loc[mnrl_selection,:]
        else:
            subset_df = self.main_df
        return subset_df

    def get_hdr_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Return a copy of the header dataframe subset of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: material collection header data
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        hdr_df = subset_df.loc[:, self.header_list]
        return hdr_df

    def get_cat_df(self,
            category: str = None,
            mineral_name: str = None) -> pd.DataFrame:
        """Returns a copy of the Categories of the array.
        Allows for selection of data from specific category and mineral type.

        :param category: categorical subset of data, defaults to None
        :type category: str, optional
        :param mineral_name: mineral subset of data, defaults to None
        :type mineral_name: str, optional
        :return: category labels
        :rtype: pd.DataFrame
        """
        subset_df = self.get_subset_df(category, mineral_name)
        cat_df = pd.DataFrame(subset_df.Category)
        return cat_df

    def get_mineral_list(self,
            category: str=None,
            unique: bool=False) -> List:
        """Get the list of mineral types in the DataFrame, and give the
        list of unique mineral types if requested

        :param category: give mineral list belonging to category only, defaults
            to None
        :type category: str, optional
        :param unique: request unique mineral names only, defaults to False
        :type unique: bool, optional
        :return: all mineral names or unique mineral names in dataframe
        :rtype: List
        """
        if category is not None:
            cat_selection = self.main_df.Category == category
            mineral_list = self.main_df.loc[cat_selection, 'Mineral Name']
        else:
            mineral_list = self.main_df['Mineral Name']
        if unique:
            mineral_list = mineral_list.unique()
        return mineral_list

    def export_main_df(self):
        """Export the dataframe to csv and pickle file formats."""

        if cfg.TIME_IT:
            tic = time.perf_counter()
        print('Exporting the Material Collection to CSV and Pickle formats...')
        table_path = Path(self.object_dir / 'tables')
        table_path.mkdir(parents=True, exist_ok=True)
        csv_out_file = Path(table_path, 'material_collection.csv')
        self.main_df.transpose().to_csv(csv_out_file)
        pkl_out_file = Path(self.object_dir, 'material_collection.pkl')
        self.main_df.to_pickle(pkl_out_file)
        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Material Collection exported in {toc - tic:0.4f} seconds.")

    def plot_profiles(self, categories_only: bool=False, ci: bool=False):
        """Plot the profiles of the materials
        """
        plotter = SpectralLibraryAnalyser(self)
        plotter.plot_profiles(categories_only=categories_only, ci=ci)



"""Spectral Library Analyser Class

Performs analysis on the given Spectral Library,
visualising and measuring the locations and strengths of
spectral features, independent of instrument resolution.

Part of the Spectral Parameters Toolkit
Author: Roger Stabbins, NHM
Date: 18-07-2022
"""
from pathlib import Path
from ast import literal_eval
import copy
import os
import time
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np
import seaborn as sns
import pysptools.spectro as spectro
import sptk.config as cfg

class SpectralLibraryAnalyser():
    """A Class for analysing the contents of a MaterialCollection or Observation
    spectral library.
    """

    def __init__(
            self,
            spectra_obj: object,
            synthetic: bool=False,
            ) -> None:
        """Create a SpectralLibraryAnalyser object

        TODO figure out type hints for Union[MaterialCollection, Observation]
        without circular imports

        :param spectra_obj: Material Collection or Observation to analyse
        :type spectra_obj: Union[MaterialCollection, Observation]
        :param synthetic: _description_, defaults to False
        :type synthetic: bool, optional
        :raises ValueError: _description_
        """
        try:
            self.obj_type = os.path.basename(spectra_obj.object_dir)
        except ValueError as exc:
            raise ValueError('Spectra object type not recognised') from exc

        self.wvls = spectra_obj.wvls
        self.project_name = spectra_obj.project_name
        self.synthetic = synthetic
        if self.synthetic:
            synthetic_tag = 'synthetic'
        else:
            synthetic_tag = 'original'
        self.project_dir = Path(
                                spectra_obj.project_dir,
                                self.obj_type,
                                'analysis',
                                synthetic_tag)
        self.project_dir.mkdir(parents=True, exist_ok=True)
        self.spectra_obj = spectra_obj
        self.spectra_obj_cr = None
        # check if the feature table has already been made
        band_file = Path(self.project_dir, 'band_info', 'band_info.csv')
        if os.path.isfile(band_file):
            self.band_info = pd.read_csv(band_file, converters={
                    'band_centres': literal_eval,
                    'band_depths': literal_eval,
                    'band_widths': literal_eval,
                    'band_areas': literal_eval})
            print(f'Band Info loaded from {band_file}')
        else:
            self.band_info = pd.DataFrame()

    def plot_profiles(self,
            with_noise: bool=False,
            scope: str='all',
            categories_only: bool=False,
            ci: bool=False):
        """Plot the profiles of the materials of the spectral library
        """
        if cfg.TIME_IT:
            tic = time.perf_counter()
            print('Plotting reflectance profiles of materials...')

        # get the reflectance data for all minerals and for each category
        refl_df = self.spectra_obj.get_refl_df()
        cat_df = self.spectra_obj.get_cat_df()
        all_df = pd.concat([refl_df, cat_df], axis=1)
        self.render_profile_plot(all_df,
                    scope=scope,
                    with_noise=with_noise,
                    ci=ci)

        if categories_only:
            if cfg.TIME_IT:
                toc = time.perf_counter()
                print(f"Reflectance profiles plotted in {toc - tic:0.4f} s.")
            return

        # plot for each category and mineral name
        for cat in self.spectra_obj.categories:
            mineral_list = self.spectra_obj.get_mineral_list(cat, unique=True)
            for mnrl in mineral_list:
                # get data for category and mineral
                refl_df = self.spectra_obj.get_refl_df(category=cat,
                                                            mineral_name=mnrl)
                cat_df = self.spectra_obj.get_cat_df(category=cat,
                                                            mineral_name=mnrl)
                cat_mnrl_df = pd.concat([refl_df, cat_df], axis=1)
                self.render_profile_plot(cat_mnrl_df,
                            cat=cat,
                            mnrl=mnrl,
                            scope=scope,
                            with_noise=with_noise,
                            ci=ci)

        if cfg.TIME_IT:
            toc = time.perf_counter()
            print(f"Reflectance profiles plotted in {toc - tic:0.4f} s.")

    def render_profile_plot(self,
            data_df: pd.DataFrame,
            cat: str='all',
            mnrl: str='entries',
            ci: bool=False,
            scope: str='all',
            with_noise: bool=False) -> None:
        """Method for producing the plot itself, according to given DataFrame,
        class, mineral name, and scope.

        :param data_df: Material Collection or Observation to DataFrame
        :type data_df: pd.DataFrame
        :param cat: Class label, defaults to 'all'
        :type cat: str, optional
        :param mnrl: Mineral Name, defaults to 'entries'
        :type mnrl: str, optional
        :param ci: Plot mean spectra with condifence interval, defaults to False
        :type ci: bool, optional
        :param scope: indicates specific elements or all data, defaults to 'all'
        :type scope: str, optional
        :param with_noise: INdicates if noise has been added, defaults to False
        :type with_noise: bool, optional
        """

        data_df = data_df.reset_index()
        data_df =pd.melt(data_df, id_vars=['Data ID','Category'])

        out_dir = Path(self.spectra_obj.object_dir / 'plots')
        out_dir.mkdir(parents=True, exist_ok=True)

        if with_noise:
            sfx = '_with_noise'
        else:
            sfx = ''

        fig_size = (1.3*cfg.FIG_SIZE[0], cfg.FIG_SIZE[1])
        fig, ax = plt.subplots(figsize=fig_size, dpi=cfg.DPI)
        y_max = max([1.0, data_df.value.max()])

        if cat == 'all':
            hue_flag = 'Category'
        else:
            hue_flag = 'Data ID'

        if self.obj_type == 'observation':
            marker_flag = True
        else:
            marker_flag = False

        plt.rcParams.update({'font.size': 8, 'lines.markersize': 3})
        ax.set(
            xbound=(cfg.SAMPLE_RES['wvl_min']-10,cfg.SAMPLE_RES['wvl_max']+10),
            ybound=(-0.1, y_max+0.1),
            autoscale_on=False)

        if ci:
            # long form version of plotting, to aggregate
            sns.lineplot(
                data=data_df,
                x='variable',
                y='value',
                hue=hue_flag,
                style=hue_flag,
                markeredgewidth=0.0,
                markers=marker_flag,
                errorbar='sd',
                lw=0.7,
                ax=ax)
        else:
            sns.lineplot(
                data=data_df,
                x='variable',
                y='value',
                hue=hue_flag,
                style=hue_flag,
                markeredgewidth=0.0,
                units='Data ID',
                estimator=None,
                lw=0.5, markers=marker_flag,
                ax=ax)

        ax.set_xlabel('Wavelength (nm)', fontsize=cfg.LABEL_S)
        ax.set_ylabel('Reflectance', fontsize=cfg.LABEL_S)
        # add minor grid lines at 50 nm intervals and major gridlines
        ax.get_xaxis().set_minor_locator(mpl.ticker.AutoMinorLocator())
        ax.grid(True, which='major',axis='both', lw=0.6)
        ax.grid(True, which='minor',axis='both', lw=0.3)

        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        # ax.legend(loc='upper right')
        plt.setp(ax.get_legend().get_texts(), fontsize=cfg.LEGEND_S)
        plt.setp(ax.get_legend().get_title(), fontsize=cfg.LEGEND_S)

        # plot title
        project_str = self.spectra_obj.project_name.replace('_', ' ')
        if self.obj_type == 'observation':
            leg_title = f'Class: {cat}, Group: {mnrl} ({scope} data) - sampled'
            if ci:
                title = 'Mean ± 1σ Sampled Spectral Library'
            else:
                title = 'Instrument Sampled Spectral Library'
        else:
            leg_title = f'Class: {cat}, Group: {mnrl} ({scope} data)'
            title = 'High-Resolution Spectral Library'
        plt.title(title, fontsize=cfg.TITLE_S)

        # save legend separately
        if cat != 'all':
            label_params = ax.get_legend_handles_labels()
            ax.get_legend().remove()
            figl, axl = plt.subplots(figsize=cfg.FIG_SIZE, dpi=cfg.DPI)
            axl.axis(False)
            n_ids = len(data_df['Data ID'].unique())
            leg = axl.legend(*label_params, loc="center",
                bbox_to_anchor=(0.5, 0.5),
                ncol=-(-n_ids // 20),
                fontsize='xx-small')
            leg.set_title(leg_title, prop={'size': 'x-small'})

        fig.tight_layout()
        # if cat != 'all':
        #     try:
        #         figl.tight_layout()
        #     except UserWarning:
        #         axl.legend(*label_params, loc="center",
        #             bbox_to_anchor=(0.5, 0.5),
        #             ncol=-(-data_df.shape[1] // 35),
        #             fontsize=3.0)

        # save figure
        project_str = self.spectra_obj.project_name
        if self.obj_type == 'observation':
            inst = self.spectra_obj.instrument.name
            filename = f'{project_str}_{inst}_{cat}_{mnrl}_{scope}'+sfx
        else:
            filename = f'{project_str}_{cat}_{mnrl}_{scope}'+sfx
        output_file = Path(out_dir, filename).with_suffix(cfg.PLT_FRMT)
        fig.savefig(output_file)
        if cat != 'all':
            legend_file=Path(out_dir,filename+'_lgnd').with_suffix(cfg.PLT_FRMT)
            figl.savefig(legend_file)

    def remove_continuum(self):
        """Remove the continuum from all spectra, and overwrite the local copy
        of the spectra object reflectance data.
        """
        self.spectra_obj_cr = copy.deepcopy(self.spectra_obj)
        cat_s = self.spectra_obj_cr.get_cat_df()
        spectra = self.spectra_obj_cr.get_refl_df()
        self.spectra_obj_cr.object_dir = Path(
                                self.project_dir,'continuum_removed')
        for index, spectrum in spectra.iterrows():
            # remove NaNs prior to analysis
            notnans = spectrum[~spectrum.isna()].index
            spectrum.index = self.wvls
            spectrum = spectrum.dropna()
            wvls = spectrum.index
            try:
                schq = spectro.SpectrumConvexHullQuotient(spectrum.tolist(),
                                                                 wvls.tolist())
                # TODO investigate: 1. why some entries are skipped,
                # TODO              2. why some hull fitting routines fail.
            except (ValueError, TypeError) as error:
                print(index)
                raise error
            path = Path(
                    self.spectra_obj_cr.object_dir,
                    'cr_algorithm_plots',
                    cat_s.loc[index].Category)
            path.mkdir(parents=True, exist_ok=True)
            schq.plot(path, index, suffix=None)
            # make sure that the correct wavelengths are added back in after
            spectra.loc[index][notnans] = schq.get_continuum_removed_spectrum()
        # rewrite spectra_obj object with continuum removed spectra
        data = spectra.reset_index()
        self.spectra_obj_cr.set_refl_data(data)
        plotter = SpectralLibraryAnalyser(self.spectra_obj_cr)
        plotter.plot_profiles()
        return self.spectra_obj_cr.main_df

    def visualise_spectrogram(self, continuum_removed: bool=True):
        """Display the reflectance data in 2D density plots, with colour giving
        absorption depth.

        :param continuum_removed: indicate to use continuum_removed data,
                defaults to True
        :type continuum_removed: bool, optional
        """
        # get reflectance sorted by Category, Mineral Name, then Sample ID
        if continuum_removed:
            try:
                vis_df = self.spectra_obj_cr.main_df
            except AttributeError:
                self.remove_continuum()
                vis_df = self.spectra_obj_cr.main_df
        vis_df.sort_values(by=['Category', 'Mineral Name', 'Sample ID'],
                                            inplace=True, ignore_index = True)

        if self.obj_type == 'material_collection':
            title = 'High-Resolution Spectral Library'
            data = vis_df[self.wvls].to_numpy() - 1.0
        elif self.obj_type == 'observation':
            title=f'{self.spectra_obj.instrument.name} sampled Spectral Library'
            data = self.spectra_obj_cr.resample_wavelengths() - 1.0
        else:
            raise ValueError("spectral object type not recognised")

        # get locations of each category
        cat_ticks = []
        cat_label_y = {}
        cat_bounds = {}
        for cat in self.spectra_obj.categories:
            cat_lo = min(vis_df[vis_df.Category == cat].index)
            cat_hi = max(vis_df[vis_df.Category == cat].index)
            cat_bounds[cat] = [cat_lo, cat_hi]
            cat_label_y[cat] = np.mean([cat_lo, cat_hi]) + 1
            cat_ticks.append(cat_lo)

        # get locations of each Mineral Name
        min_ticks = []
        mineral_names = vis_df['Mineral Name'].unique()
        mineral_label_y = {}
        mineral_bounds = {}
        for mineral_name in mineral_names:
            min_lo = min(vis_df[vis_df['Mineral Name'] == mineral_name].index)
            min_hi = max(vis_df[vis_df['Mineral Name'] == mineral_name].index)
            mineral_bounds[mineral_name] = [min_lo, min_hi]
            mineral_label_y[mineral_name] = np.mean([min_lo, min_hi])
            min_ticks.append(min_lo)

        # set plot limits
        wvl_lo = cfg.SAMPLE_RES['wvl_min']
        wvl_hi = cfg.SAMPLE_RES['wvl_max']
        # set up a good figure size so that good # of samples are shown per cm.
        # A4 = 210 x 297 mm
        # minus 3 cm for border
        # fig size - width = 190 mm
        # fig size - height = 2mm * #samples
        height = 0.2*len(data)
        if height < 15:
            height = 15 # limit the minimum height to 15 cm
        fig, ax = plt.subplots(figsize=[15*cfg.CM, height*cfg.CM], dpi=cfg.DPI)
        cmap = plt.get_cmap('viridis_r')
        cmap.set_bad('black')

        # draw plot
        im = ax.imshow(
                data,
                aspect='auto',
                extent=[wvl_lo, wvl_hi, 0, len(data)],
                interpolation='nearest',
                origin='lower',
                cmap=cmap, vmin=-1.0, vmax=0.0)
        ax.set_xlabel('Wavelength (nm)')
        ax.set_title(title)

        # add mineral name labels
        for mineral_name in mineral_names:
            ax.annotate(
                '',
                xy=(wvl_hi + 50, mineral_bounds[mineral_name][0]),
                xytext=(wvl_hi + 50, mineral_bounds[mineral_name][1]+1),
                arrowprops=dict(arrowstyle='<|-|>', shrinkA=0, shrinkB=0),
                annotation_clip=False)
            ax.annotate(
                mineral_name,
                xy=(wvl_hi + 70,mineral_label_y[mineral_name]),
                xytext=(wvl_hi + 70,mineral_label_y[mineral_name]),
                rotation=45,
                ha='left',
                va='bottom',
                annotation_clip=False)

        # add category labels
        for cat in self.spectra_obj.categories:
            ax.annotate(
                '',
                xy=(wvl_lo - 50, cat_bounds[cat][0]),
                xytext=(wvl_lo - 50, cat_bounds[cat][1]+1),
                arrowprops=dict(arrowstyle='<|-|>', shrinkA=0, shrinkB=0),
                annotation_clip=False)
            ax.annotate(
                cat,
                xy=(wvl_lo - 70,cat_label_y[cat]),
                xytext=(wvl_lo - 70,cat_label_y[cat]),
                rotation=45,
                ha='right',
                va='top',
                annotation_clip=False)

        # set ticks
        minor_ticks = np.arange(0, len(data)+1, 1)
        ax.set_yticks(minor_ticks, minor=True)
        ax.grid(which='minor', axis='y', lw=0.5)
        ax.set_yticks(min_ticks, minor=False)
        ax.tick_params(right=True)
        ax.set_yticklabels([])
        ax.grid(which='major', axis='y', lw=0.8, color='w')
        ax.tick_params(right=True)

        divider = make_axes_locatable(ax)
        cax = divider.new_vertical(size=0.5*cfg.CM, pad=0.6, pack_start = True)
        fig.add_axes(cax)

        cbar = plt.colorbar(im, cax = cax, orientation='horizontal', shrink=0.5)
        cbar.set_label('Band Depth')

        # export
        filepath=Path(self.project_dir,'spectrogram').with_suffix(cfg.PLT_FRMT)
        plt.savefig(filepath, bbox_inches='tight', pad_inches = 0.1)

    # """
    # Methods still in development
    # """

    def analyse_bands(self):
        """Find the centre-wavelengths, fwhms, depths and areas of distinct
        bands in each entry.
        Note: method is not complete
        Note: this is not a time sensitive operation, so have not attempted to
        parallelise.
        """

        print('Extracing Band Info...')
        cat_s = self.spectra_obj.get_cat_df()
        spectra = self.spectra_obj.get_refl_df()
        # prepare containers for feature information
        indices = []
        n_features = []
        band_centres = []
        band_depths = []
        band_widths = []
        band_areas = []
        for index, spectrum in spectra.iterrows():
            path = Path(self.project_dir, 'features', cat_s.loc[index].Category)
            path.mkdir(parents=True, exist_ok=True)
            # remove NaNs prior to analysis
            spectrum.index = self.wvls
            spectrum = spectrum.dropna()
            wvls = spectrum.index
            try:
                fea = spectro.FeaturesConvexHullQuotient(
                            spectrum.tolist(),
                            wvls.tolist(),
                            baseline=0.98)
                fea.plot(path, index, feature='all')
                n_feat = fea.get_number_of_kept_features()
                bcs = [fea.get_absorbtion_wavelength(f)
                                                    for f in range(0, n_feat)]
                bds = [1.0 - fea.get_absorbtion_depth(f)
                                                    for f in range(0, n_feat)]
                bws = [fea.get_full_width_at_half_maximum(f)
                                                    for f in range(0, n_feat)]
                bas = [fea.get_area(f) for f in range(0, n_feat)]

                # sort the lists by band centres (lowest to highest)
                bds = [x for _, x in sorted(zip(bcs, bds))]
                bws = [x for _, x in sorted(zip(bcs, bws))]
                bas = [x for _, x in sorted(zip(bcs, bas))]
            except TypeError:
                n_feat = []
                bcs = []
                bds = []
                bws = []
                bas = []

            # append to make list of lists
            indices.append(index)
            n_features.append([n_feat])
            band_centres.append(sorted(bcs))
            band_depths.append(bds)
            band_widths.append(bws)
            band_areas.append(bas)
            print(f"{index}: {n_feat} features")

        # put the feature information in a DataFrame
        df = pd.DataFrame(
                data = {
                    'n_features'  : n_features,
                    'band_centres': band_centres,
                    'band_depths' : band_depths,
                    'band_widths' : band_widths,
                    'band_areas'  : band_areas},
                index = indices)
        hdr_df = self.spectra_obj.get_hdr_df() # add metadata for entries

        # put this as a new feature of the object
        self.band_info = pd.concat([hdr_df, df], axis=1)

        return self.band_info

    def export_band_info(self):
        """Export band info to csv
        """

        path = Path(
            self.project_dir,
            'features', 'feature_table').with_suffix('.csv')
        self.band_info.sort_values(
            by=['Category', 'Mineral Name', 'Sample ID'],
            inplace=True,
            ignore_index = True)
        self.band_info.to_csv(path)

        # Order the data for export
        band_centre_info = [
            'Category',
            'Mineral Name',
            'band_centres',
            'band_widths',
            'band_depths',
            'band_areas']
        bc_df = self.band_info.loc[:,band_centre_info]

        bc_df['Band Centres (nm)'] = [
            ', '.join(map("{:.0f}".format, l)) for l in bc_df['band_centres']]
        bc_df.drop('band_centres', axis=1, inplace=True)
        bc_df['Band Widths (nm)'] = [
            ', '.join(map("{:.0f}".format, l)) for l in bc_df['band_widths']]
        bc_df.drop('band_widths', axis=1, inplace=True)
        bc_df['Band Depths'] = [
            ', '.join(map("{:.2f}".format, l)) for l in bc_df['band_depths']]
        bc_df.drop('band_depths', axis=1, inplace=True)
        bc_df['Band Areas'] = [
            ', '.join(map("{:.1f}".format, l)) for l in bc_df['band_areas']]
        bc_df.drop('band_areas', axis=1, inplace=True)

        path = Path(
            self.project_dir, 'features', 'feature_table').with_suffix('.tex')
        bc_df.to_latex(path, float_format="%.3f", longtable=True,
                       column_format='lll|p{3.5cm}|p{3.5cm}|p{3.5cm}|p{3.5cm}')

        # Use this script to generate latex table:
        # \documentclass[a4paper, landscape]{article}
        # \usepackage[a4paper,margin=1in,landscape]{geometry}
        # \usepackage{longtable}
        # \usepackage{booktabs}
        # \begin{document}
        # \include{feature_table}
        # \end{document}

        print('end')

    def synthesize_spectra_from_band_info(self) -> object:
        """Produce a new MaterialCollection with reflectance spectra synthesized
        from the band info extracted. The idea is to compare the visualisation
        of this synthesized spectra with the original input data.

        :return: A copy of the Material Collection with synthesized spectra
        :rtype: MaterialCollection
        """

        print('begin synthesizing band spectra')
        mat_synthetic = copy.deepcopy(self.spectra_obj)
        spectra = mat_synthetic.get_refl_df()
        for _, entry in self.band_info.iterrows():
            print(entry['Data ID'])
            spectrum = np.zeros(len(self.wvls))
            cwl = np.array(entry['band_centres'])
            fwhm = np.array(entry['band_widths'])
            depth = np.array(entry['band_depths'])
            if len(cwl) != 0:
                gauss = self.build_gauss_feature(cwl, fwhm, depth)
                spectrum+=gauss
            spectrum = -spectrum + 1
            spectra.loc[entry['Data ID']] = spectrum

        # overwrite reflectance with spectra array
        data = spectra.reset_index()
        mat_synthetic.set_refl_data(data)
        mat_synthetic.project_dir = Path(
                                self.spectra_obj.project_dir,
                                self.obj_type,
                                'analysis',
                                'synthetic')
        mat_synthetic.plot_material_profiles()
        # return new Spectral Library
        return mat_synthetic

    def build_gauss_feature(self,
            cwl: np.array,
            fwhm: np.array,
            depth: np.array) -> np.array:
        """Build Gaussian absorption feature profile according to the given cwl,
        fwhm, depths and wvls.

        :param cwl: Centre wavelength(s) (nm)
        :type cwl: np.array
        :param fwhm: Full-Width at Half Maximum(s) (nm)
        :type fwhm: np.array
        :param depth: Absorption feature depth(s)
        :type depth: np.array
        :returns: Gaussian absorption profile
        :rtype: np.array
        """
        sig = fwhm / 2.355482004503 # convert from fwhm to 1-sigma
        # vectorisation: extend cwls, sigs & wvls to match dimensions
        cwls = np.tile(cwl, [self.wvls.shape[0],1])
        sigs = np.tile(sig, [self.wvls.shape[0],1])
        depths = np.tile(depth, [self.wvls.shape[0],1])
        wvls = np.tile(self.wvls, [cwl.shape[0],1]).transpose()
        # compute the Gaussian profiles in parallel
        gauss = depths*np.exp(-np.power(wvls-cwls,2.)/(2*np.power(sigs, 2.)))
        gauss = np.sum(gauss, axis=1)
        # sum over the correct axis to get final profile
        return gauss
